import torch
import torch.nn as nn
import torch.nn.functional as F


class GMUC_Loss(nn.Module):
    """GMUC Loss

    Attributes:
        args: Some pre-set parameters, etc
        model: The UKG model for training.
    """
    def __init__(self, args, model):
        super(GMUC_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, query_scores, query_scores_var, query_ae_loss, false_scores, query_confidence):
        """
        Args:
            query_scores: The matching scores for link prediction.
            query_scores_var: The prediction value of confidence.
            query_ae_loss: The loss in the matching process,
            false_scores: The loss for false set.
            query_confidence: The true value for confidence in query set.
        Returns:
            loss: The training loss for back propagation.
        """
        if self.args.num_neg != 1:
            false_scores = false_scores.reshape((query_scores.shape[0], self.args.num_neg))  # resize
            false_scores = torch.mean(false_scores, dim=1)
            false_scores = false_scores.reshape((query_scores.shape[0]))  # resize

        zero_torch = torch.zeros(query_confidence.shape).cuda()
        query_conf_mask = torch.where(query_confidence < 0.5, zero_torch, query_confidence)
        # ------ MSE loss -------
        mae_loss = (query_scores_var - query_confidence) ** 2
        mae_loss = self.args.mae_weight * mae_loss.sum()
        # ------ rank loss ------
        rank_loss = self.args.margin - (query_scores - false_scores)
        if self.args.if_conf:
            rank_loss = torch.mean(F.relu(rank_loss) * query_conf_mask)  # rank loss
        else:
            rank_loss = torch.mean(F.relu(rank_loss))
        rank_loss = self.args.rank_weight * rank_loss
        # ------ lstem loss ------
        ae_loss = self.args.ae_weight * query_ae_loss  # lstm aggregation loss
        # ------ over all loss ------
        loss = rank_loss + mae_loss + ae_loss

        return loss



import torch


def link_predict(batch, model, prediction="all"):
    """The evaluate task is predicting the head entity or tail entity in incomplete triples.

    Args:
        batch: The batch of the triples for validation or test.
        model: The UKG model for training.
        predicion: mode of link prediction.

    Returns:
        ranks: The rank of the triple to be predicted.
    """
    if prediction == "all":
        tail_ranks = tail_predict(batch, model)
        head_ranks = head_predict(batch, model)
        ranks = torch.cat([tail_ranks, head_ranks])
    elif prediction == "head":
        ranks = head_predict(batch, model)
    elif prediction == "tail":
        ranks = tail_predict(batch, model)

    return ranks.float()


def link_predict_filter(batch, model, confidence, prediction="all"):
    """The evaluate task is predicting the head entity or tail entity in incomplete triples.
    link_predict_filter is for high confidence test samples,
    only samples higher than the set confidence will participate in link prediction tasks.
    When confidence is set to 0, this function is the same as link_predict.

    Args:
        batch: The batch of the triples for validation or test.
        model: The UKG model for training.
        predicion: mode of link prediction.

    Returns:
        ranks: The rank of the triple to be predicted.
    """
    if prediction == "all":
        tail_ranks = tail_predict_filter(batch, model)
        head_ranks = head_predict_filter(batch, model)
        ranks = torch.cat([tail_ranks, head_ranks])
    elif prediction == "head":
        ranks = head_predict_filter(batch, model, confidence)
    elif prediction == "tail":
        ranks = tail_predict_filter(batch, model, confidence)

    return ranks.float()


def head_predict_filter(batch, model, confidence):
    """Getting head entity ranks (high confidence).

    Args:
        batch: The batch of the triples for validation or test
        model: The UKG model for training.

    Returns:
        tensor: The rank of the head entity to be predicted, dim [batch_size]
    """
    pos_triple = batch["positive_sample"]
    mask = pos_triple[:, -1] >= confidence
    pos_triple_filtered = pos_triple[mask]
    batch_copy = batch.copy()
    batch_copy["positive_sample"] = pos_triple_filtered
    idx = pos_triple_filtered[:, 0].long()
    label = batch["head_label"].clone()
    label = label[mask]
    pred_score = model.get_score(batch_copy, "head_predict")
    return calc_ranks(idx, label, pred_score)


def tail_predict_filter(batch, model, confidence):
    """Getting tail entity ranks.(high confidence)

    Args:
        batch: The batch of the triples for validation or test
        model: The UKG model for training.

    Returns:
        tensor: The rank of the tail entity to be predicted, dim [batch_size]
    """
    pos_triple = batch["positive_sample"]
    mask = pos_triple[:, -1] >= confidence
    pos_triple_filtered = pos_triple[mask]
    batch_copy = batch.copy()
    batch_copy["positive_sample"] = pos_triple_filtered
    idx = pos_triple_filtered[:, 2].long()
    label = batch["tail_label"].clone()
    label = label[mask]
    pred_score = model.get_score(batch_copy, "tail_predict")
    return calc_ranks(idx, label, pred_score)

def link_predict_raw(batch, model, confidence, prediction="all"):
    """The evaluate task is predicting the head entity or tail entity in incomplete triples.
    Only samples higher than the set confidence will participate in link prediction tasks.
    Different from link_predict_filter, this function does not filter samples that already exist in the knowledge graph when calculating the ranking of samples

    Args:
        batch: The batch of the triples for validation or test.
        model: The KG model for training.
        predicion: mode of link prediction.

    Returns:
        ranks: The rank of the triple to be predicted.
    """
    if prediction == "all":
        tail_ranks = tail_predict_raw(batch, model)
        head_ranks = head_predict_raw(batch, model)
        ranks = torch.cat([tail_ranks, head_ranks])
    elif prediction == "head":
        ranks = head_predict_raw(batch, model, confidence)
    elif prediction == "tail":
        ranks = tail_predict_raw(batch, model, confidence)

    return ranks.float()

def head_predict_raw(batch, model, confidence):
    """Getting head entity ranks. (high confidence + no filter)

    Args:
        batch: The batch of the triples for validation or test
        model: The UKG model for training.

    Returns:
        tensor: The rank of the head entity to be predicted, dim [batch_size]
    """
    pos_triple = batch["positive_sample"]
    mask = pos_triple[:, -1] >= confidence
    pos_triple_filtered = pos_triple[mask]
    batch_copy = batch.copy()
    batch_copy["positive_sample"] = pos_triple_filtered
    idx = pos_triple_filtered[:, 0].long()
    label = batch["head_label"].clone()
    label = label[mask]
    pred_score = model.get_score(batch_copy, "head_predict")
    return calc_ranks_raw(idx, label, pred_score)


def tail_predict_raw(batch, model, confidence):
    """Getting tail entity ranks. (high confidence + no filter)

    Args:
        batch: The batch of the triples for validation or test
        model: The UKG model for training.

    Returns:
        tensor: The rank of the tail entity to be predicted, dim [batch_size]
    """
    pos_triple = batch["positive_sample"]
    mask = pos_triple[:, -1] >= confidence
    pos_triple_filtered = pos_triple[mask]
    batch_copy = batch.copy()
    batch_copy["positive_sample"] = pos_triple_filtered
    idx = pos_triple_filtered[:, 2].long()
    label = batch["tail_label"].clone()
    label = label[mask]
    pred_score = model.get_score(batch_copy, "tail_predict")
    return calc_ranks_raw(idx, label, pred_score)

def head_predict(batch, model):
    """Getting head entity ranks.

    Args:
        batch: The batch of the triples for validation or test
        model: The UKG model for training.

    Returns:
        tensor: The rank of the head entity to be predicted, dim [batch_size]
    """
    pos_triple = batch["positive_sample"]
    idx = pos_triple[:, 0].long()
    label = batch["head_label"]
    pred_score = model.get_score(batch, "head_predict")
    return calc_ranks(idx, label, pred_score)


def tail_predict(batch, model):
    """Getting tail entity ranks.

    Args:
        batch: The batch of the triples for validation or test
        model: The UKG model for training.

    Returns:
        tensor: The rank of the tail entity to be predicted, dim [batch_size]
    """
    pos_triple = batch["positive_sample"]
    idx = pos_triple[:, 2].long()
    label = batch["tail_label"]
    pred_score = model.get_score(batch, "tail_predict")
    return calc_ranks(idx, label, pred_score)


def calc_ranks(idx, label, pred_score):
    """Calculating triples score ranks.

    Args:
        idx ([type]): The id of the entity to be predicted.
        label ([type]): The id of existing triples, to calc filtered results.
        pred_score ([type]): The score of the triple predicted by the model.

    Returns:
        ranks: The rank of the triple to be predicted, dim [batch_size].
    """
    b_range = torch.arange(pred_score.size()[0])  # get batch_size

    target_pred = pred_score[b_range, idx]  #  idx is the real tail entity number in UKG
    # Set the score of the sample with label 1 to be very low and filter it to exclude samples that are already in the knowledge graph
    pred_score = torch.where(label.bool(), -torch.ones_like(pred_score) * 10000000, pred_score)
    pred_score[b_range, idx] = target_pred

    # Get the ranking of each score in the pred_score.
    ranks = (
            1
            + torch.argsort(
        torch.argsort(pred_score, dim=1, descending=True), dim=1, descending=False
    )[b_range, idx]
    )
    return ranks

def calc_ranks_raw(idx, label, pred_score):
    """Calculating triples score ranks.
    Different from calc_ranks, calc_ranks_raw will not exclude samples that are already in the knowledge graph when get the ranks.

    Args:
        idx ([type]): The id of the entity to be predicted.
        label ([type]): The id of existing triples, to calc filtered results.
        pred_score ([type]): The score of the triple predicted by the model.

    Returns:
        ranks: The rank of the triple to be predicted, dim [batch_size].
    """
    b_range = torch.arange(pred_score.size()[0])  # 得到batch_size

    target_pred = pred_score[b_range, idx]   #  idx is the real tail entity number in UKG

    pred_score[b_range, idx] = target_pred

    # Get the ranking of each score in the pred_score.
    ranks = (
            1
            + torch.argsort(
        torch.argsort(pred_score, dim=1, descending=True), dim=1, descending=False
    )[b_range, idx]
    )
    return ranks




import torch.nn as nn
import torch
from .model import Model
from gensim.models.word2vec import LineSentence
from gensim.models import Word2Vec
import torch.optim as optim
import numpy as np
import multiprocessing
import time
import os
os.environ['CUDA_LAUNCH_BLOCKING'] = '1'


class LSTMModel(nn.Module):
    """
    Create LSTM model.
    """

    def __init__(self, dim):
        super(LSTMModel, self).__init__()
        self.lstm = nn.LSTM(dim, hidden_size=dim, batch_first=True)
        self.fc = nn.Linear(dim, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        lstm_output, _ = self.lstm(x)
        output = self.fc(lstm_output[:, -1, :])
        output = self.sigmoid(output)
        return output   # The shape is [bs, 1]，and every elment represent the probability of the corresponding sample's result, within range [0,1]


class UKGsE(Model):

    def __init__(self, args):
        super(UKGsE, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None

        self.lstm_model = LSTMModel(args.emb_dim).to(args.gpu)

        self.init_emb()

    def init_emb(self):
        """
        Pre-train the entity and relation embeddings using Word2Vec model.
        """
        self.ready_data()

        corpus = self.args.data_path + '/train.tsv.txt'    # relation is represented in the form of r+relation_id

        word2vec_model = Word2Vec(LineSentence(corpus), vector_size=int(self.args.emb_dim), window=2, sample=0,
                                  epochs=10, negative=5, min_count=1,
                                  sg=self.args.sg,  # skip_gram (1) or CBOW (0)
                                  workers=multiprocessing.cpu_count())
        word2vec_model.init_sims(replace=True)
        # word2vec_model.wv.save_word2vec_format(vectors)

        # Initialize the embedding layer of entity and relation
        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)

        # Assign the embeddings of entity and relation using the word vectors of Word2Vec model
        for i in range(self.args.num_ent):
            ent_word = str(i)  # Construct the corresponding word according to the entity's id
            if ent_word in word2vec_model.wv:
                ent_vector = torch.Tensor(word2vec_model.wv[ent_word].copy()).unsqueeze(0)
                self.ent_emb.weight.data[i] = ent_vector  # Assign word vectors to numbered entity vectors

        for i in range(self.args.num_rel):
            rel_word = "r" + str(i)  # Construct the corresponding word according to the relation's id
            if rel_word in word2vec_model.wv:
                rel_vector = torch.Tensor(word2vec_model.wv[rel_word].copy()).unsqueeze(0)
                self.rel_emb.weight.data[i] = rel_vector

        self.ent_emb = self.ent_emb.to(self.args.gpu)
        self.rel_emb = self.rel_emb.to(self.args.gpu)

    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        
        if mode == 'head_predict' and head_emb.shape[0] == 1:
            # print('head_predict')
            head_emb = head_emb.squeeze(0)  # [1, num_ent, dim]-->[num_ent, dim]
            relation_emb = relation_emb.squeeze(1)  # [bs, 1, dim]-->[bs, dim]
            tail_emb = tail_emb.squeeze(1)  # [bs, 1, dim]-->[bs, dim]
            X = torch.empty(head_emb.shape[0] * tail_emb.shape[0], 3, tail_emb.shape[1])  # [bs*num_ent, 3, dim]

            tail_emb_repeat = tail_emb.repeat_interleave(head_emb.shape[0], dim=0)
            relation_emb_repeat = relation_emb.repeat_interleave(head_emb.shape[0], dim=0)

            head_emb_repeat = head_emb.repeat(tail_emb.shape[0], 1)
            head_emb_repeat = head_emb_repeat.view(-1, head_emb.shape[1])

            X[:, 0, :] = head_emb_repeat
            X[:, 1, :] = relation_emb_repeat
            X[:, 2, :] = tail_emb_repeat
            # print(X.shape)

        elif mode == 'tail_predict' and tail_emb.shape[0] == 1:  # tail_predict
            # print('tail_predict')
            head_emb = head_emb.squeeze(1)  # [bs, 1, dim]-->[bs, dim]
            relation_emb = relation_emb.squeeze(1)  # [bs, 1, dim]-->[bs, dim]
            tail_emb = tail_emb.squeeze(0)  # [1, num_ent, dim]-->[num_ent, dim]

            X = torch.empty(head_emb.shape[0] * tail_emb.shape[0], 3, head_emb.shape[1])

            head_emb_repeat = head_emb.repeat_interleave(tail_emb.shape[0], dim=0)
            relation_emb_repeat = relation_emb.repeat_interleave(tail_emb.shape[0], dim=0)

            tail_emb_repeat = tail_emb.repeat(head_emb.shape[0], 1)
            tail_emb_repeat = tail_emb_repeat.view(-1, tail_emb.shape[1])

            X[:, 0, :] = head_emb_repeat
            X[:, 1, :] = relation_emb_repeat
            X[:, 2, :] = tail_emb_repeat
            # print(X.shape)

        else:
            head_emb = head_emb.squeeze(1)
            relation_emb = relation_emb.squeeze(1)
            tail_emb = tail_emb.squeeze(1)  # [bs, dim]
            X = torch.stack((head_emb, relation_emb, tail_emb), dim=1)

        X = X.to(self.args.gpu)
        y = self.lstm_model(X)
        # print(y.shape)
        score = y.to(self.args.gpu)
        batch_size = relation_emb.shape[0]
        score = score.reshape(batch_size, -1)  

        return score

    def forward(self, triples, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        # self.lstm()

        triples = triples[:, :3].to(torch.int)
        # c = triples[:, 3]
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)  # [bs, 1, dim]
        head_emb = head_emb.squeeze(1)
        relation_emb = relation_emb.squeeze(1)
        tail_emb = tail_emb.squeeze(1)   # [bs, dim]

        X_train = torch.stack((head_emb, relation_emb, tail_emb), dim=1).to(self.args.gpu)  # [bs, 3, dim]

        score = self.lstm_model(X_train)

        return score

    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        triples = batch["positive_sample"]
        triples = triples[:, :3].to(torch.int)
        triples = triples.to(self.args.gpu)

        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def ready_data(self):
        """
        The relationships and entities in the file are numbered with numbers, and there are duplicates, so the relationship ID is preceded by the prefix 'r'
        """
        base = self.args.data_path

        train_file = base + '/train.tsv'
        test_file = base + '/test.tsv'

        train_with_confidence_file = open(train_file + ".txt", 'w')
        test_with_confidence_file = open(test_file + ".txt", 'w')

        with open(train_file, 'r') as f1, open(test_file, 'r') as f2:
            lines = f1.readlines()
            for line in lines:
                triplet = line.strip().split()
                train_with_confidence_file.write(triplet[0] + '\t' + 'r' + triplet[1] + '\t'
                                                 + triplet[2] + '\n')  # for step_1

            lines = f2.readlines()
            for line in lines:
                triplet = line.strip().split()
                test_with_confidence_file.write(triplet[0] + '\t' + 'r' + triplet[1] + '\t'
                                                + triplet[2] + '\t' + triplet[3] + '\n')

        train_with_confidence_file.close()
        test_with_confidence_file.close()




from .Adv_Loss import Adv_Loss
from .UKGE_Loss import UKGE_Loss
from .BEUrRE_Loss import BEUrRE_Loss
from .FocusE_Loss import FocusE_Loss
from .GMUC_Loss import GMUC_Loss
from .GMUCp_Loss import GMUCp_Loss
from .UKGE_PSL_Loss import UKGE_PSL_Loss
from .PASSLEAF_Loss import PASSLEAF_Loss
from .UKGsE_Loss import UKGsE_Loss
from .UPGAT_Loss import UPGAT_Loss
from .GTransE_Loss import GTransE_Loss


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch.autograd import Variable
import numpy as np

torch.set_printoptions(sci_mode=False)


class GMUCp(nn.Module):
    """`Incorporating Uncertainty of Entities and Relations into Few-Shot Uncertain Knowledge Graph Embedding`_ (GMUC+),
        which incorporates the inherent uncertainty of entities and relations into uncertain knowledge graph embedding.

    Attributes:
        args: Model configuration parameters.
        num_symbols: The sum of the number of entities and relationships.
        embed: Pretrained embedding.

    .. _Incorporating Uncertainty of Entities and Relations into Few-Shot Uncertain Knowledge Graph Embedding:
        https://link.springer.com/chapter/10.1007/978-981-19-7596-7_2
    """

    def __init__(self, args, num_symbols, embed=None):
        super(GMUCp, self).__init__()
        self.args = args
        self.num_symbols = num_symbols
        self.pad_idx = num_symbols
        # mean&variance embeddings of symbols(ents+rels)
        self.symbol_emb_mean = nn.Embedding(num_symbols + 1, self.args.emb_dim,
                                            padding_idx=num_symbols)  # initialize as N(0, 1)
        self.symbol_emb_var = nn.Embedding(num_symbols + 1, self.args.emb_dim, padding_idx=num_symbols)

        # neighbor encoder attention
        self.neigh_att_W_mean = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.neigh_att_u_mean = nn.Linear(self.args.emb_dim, 1)
        self.neigh_att_W_var = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.neigh_att_u_var = nn.Linear(self.args.emb_dim, 1)

        # aggregate support set
        self.set_rnn_encoder = nn.LSTM(2 * self.args.emb_dim, 2 * self.args.emb_dim, 1, bidirectional=False)
        self.set_rnn_decoder = nn.LSTM(2 * self.args.emb_dim, 2 * self.args.emb_dim, 1, bidirectional=False)
        # aggregation attention
        self.set_att_W_mean = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.set_att_u_mean = nn.Linear(self.args.emb_dim, 1)
        self.set_att_W_var = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.set_att_u_var = nn.Linear(self.args.emb_dim, 1)

        # matchnet variance score
        self.w = nn.Parameter(torch.randn(1), requires_grad=True)
        self.b = nn.Parameter(torch.randn(1), requires_grad=True)

        # ic_loss function
        self.ic_loss_w = nn.Parameter(torch.randn(1), requires_grad=True)
        self.ic_loss_b = nn.Parameter(torch.randn(1), requires_grad=True)
        self.ic_loss_W = nn.Linear(self.args.emb_dim, 1)

        init.xavier_normal_(self.neigh_att_W_mean.weight)
        init.xavier_normal_(self.neigh_att_W_var.weight)
        init.xavier_normal_(self.neigh_att_u_mean.weight)
        init.xavier_normal_(self.neigh_att_u_var.weight)
        init.xavier_normal_(self.set_att_W_mean.weight)
        init.xavier_normal_(self.set_att_W_var.weight)
        init.xavier_normal_(self.set_att_u_mean.weight)
        init.xavier_normal_(self.set_att_u_var.weight)

        self.softmax = nn.Softmax(dim=1)
        self.layer_norm = LayerNormalization(2 * self.args.emb_dim)
        dropout = self.args.dropout
        self.dropout = nn.Dropout(dropout)

        self.support_encoder_mean = SupportEncoder(2 * self.args.emb_dim, 2 * 2 * self.args.emb_dim, dropout)
        self.support_encoder_var = SupportEncoder(2 * self.args.emb_dim, 2 * 2 * self.args.emb_dim, dropout)
        self.matchnet_mean = MatchNet(2 * self.args.emb_dim, self.args.process_steps)
        self.matchnet_var = MatchNet(2 * self.args.emb_dim, self.args.process_steps)

    def neighbor_encoder(self, connections, num_neighbors):
        """Mean and variance encoding using node neighbors.

        Args:
            connections: Neighbor information.
            num_neighbors: The number of neighbors.

        Returns:
            out_mean: Mean embeddings.
            out_var: Variance embeddings.
        """
        num_neighbors = num_neighbors.unsqueeze(1)
        relations = connections[:, :, 0].squeeze(-1).long()
        entities = connections[:, :, 1].squeeze(-1).long()
        confidences = connections[:, :, 2]

        batch_size, neighbor_size = confidences.shape
        confidences = confidences.reshape((batch_size, neighbor_size, 1)).repeat((1, 1, 2 * self.args.emb_dim))

        # mean
        rel_embeds_mean = self.dropout(self.symbol_emb_mean(relations))
        ent_embeds_mean = self.dropout(self.symbol_emb_mean(entities))
        concat_embeds_mean = torch.cat((rel_embeds_mean, ent_embeds_mean), dim=-1)
        # variance
        rel_embeds_var = self.dropout(self.symbol_emb_var(relations))
        ent_embeds_var = self.dropout(self.symbol_emb_var(entities))
        concat_embeds_var = torch.cat((rel_embeds_var, ent_embeds_var), dim=-1)

        # attention mechanism  to enhance reps
        # mean
        out_mean = self.neigh_att_W_mean(concat_embeds_mean).tanh()
        att_w_mean = self.neigh_att_u_mean(out_mean)
        att_w_mean = self.softmax(att_w_mean).view(concat_embeds_mean.size()[0], 1, 30)
        out_mean = torch.bmm(att_w_mean, ent_embeds_mean).view(concat_embeds_mean.size()[0], self.args.emb_dim)
        # variance
        out_var = self.neigh_att_W_var(concat_embeds_var).tanh()
        att_w_var = self.neigh_att_u_var(out_var)
        att_w_var = self.softmax(att_w_var).view(concat_embeds_var.size()[0], 1, 30)
        out_var = torch.bmm(att_w_var, ent_embeds_mean).view(concat_embeds_var.size()[0], self.args.emb_dim)

        return out_mean.tanh(), out_var.tanh()

    def score_func(self, support, support_meta, query, query_meta, if_ne=True):
        """Calculating the score of query triples with support set.

        Args:
            support: Support set.
            support_meta: Neighbor information of the head and tail entities in support set.
            query: Query set.
            query_meta: Neighbor information of the head and tail entities in query set.
            if_ne: Whether to use neighbor node information for encoding.

        Returns:
            rank_score: The score for link prediction.
            conf_score: The confidence prediction value.
        """
        if not if_ne:
            support_left, support_right = support[:, 0].squeeze(-1).long(), support[:, 1].squeeze(-1).long()
            support_left_mean, support_right_mean = self.symbol_emb_mean(support_left), self.symbol_emb_mean(
                support_right)
            support_left_var, support_right_var = self.symbol_emb_var(support_left), self.symbol_emb_var(support_right)

            query_left, query_right = query[:, 0].squeeze(-1).long(), query[:, 1].squeeze(-1).long()
            query_left_mean, query_right_mean = self.symbol_emb_mean(query_left), self.symbol_emb_mean(query_right)
            query_left_var, query_right_var = self.symbol_emb_var(query_left), self.symbol_emb_var(query_right)

        else:
            query_left_connections, query_left_degrees, query_right_connections, query_right_degrees = query_meta
            support_left_connections, support_left_degrees, support_right_connections, support_right_degrees = support_meta
            # ========================== neighbor encoder ===============================
            query_left_mean, query_left_var = self.neighbor_encoder(query_left_connections, query_left_degrees)
            query_right_mean, query_right_var = self.neighbor_encoder(query_right_connections, query_right_degrees)
            support_left_mean, support_left_var = self.neighbor_encoder(support_left_connections, support_left_degrees)
            support_right_mean, support_right_var = self.neighbor_encoder(support_right_connections,
                                                                          support_right_degrees)

        # query encode
        query_neighbor_mean = torch.cat((query_left_mean, query_right_mean), dim=-1)
        query_g_mean = self.support_encoder_mean(query_neighbor_mean)
        query_neighbor_var = torch.cat((query_left_var, query_right_var), dim=-1)
        query_g_var = self.support_encoder_var(query_neighbor_var)

        # support set
        support_neighbor_mean = torch.cat((support_left_mean, support_right_mean), dim=-1)
        support_g_mean = self.support_encoder_mean(support_neighbor_mean)
        support_neighbor_var = torch.cat((support_left_var, support_right_var), dim=-1)
        support_g_var = self.support_encoder_var(support_neighbor_var)

        # =============================== matching networks==============================
        query_mean = query_g_mean
        query_var = query_g_var
        # support set mean pooling
        support_mean = torch.mean(support_g_mean, dim=0, keepdim=True)
        support_var = torch.mean(support_g_var, dim=0, keepdim=True)

        # match
        matching_scores_mean = self.matchnet_mean(support_mean, query_mean)
        matching_scores_var = self.matchnet_var(support_var, query_var)

        rank_score = matching_scores_mean + self.args.conf_score_weight * matching_scores_var
        conf_score = torch.sigmoid(self.w * matching_scores_var + self.b)
        return rank_score, conf_score

    def forward(self, support, support_meta, query, query_meta, false, false_meta, if_ne=True):
        """The functions used in the training phase.

        Args:
            support: Support set.
            support_meta: Neighbor information of the head and tail entities in support set.
            query: Query set.
            query_meta: Neighbor information of the head and tail entities in query set.
            false: False set.
            false_meta: Neighbor information of the head and tail entities in false set.
            if_ne: Whether to use neighbor node information for encoding.

        Returns:
            query_scores: The match score between the query set and the support set.
            query_scores_var: Support set confidence prediction values.
            false_scores: The match score between the false set and the support set.
            query_confidence: The confidence for query set.
        """
        query_scores, query_scores_var = self.score_func(support, support_meta, query, query_meta, if_ne)
        false_scores, false_scores_var = self.score_func(support, support_meta, false, false_meta, if_ne)
        query_confidence = query[:, 2]

        return query_scores, query_scores_var, false_scores, query_confidence


class LayerNormalization(nn.Module):
    """
    Layer normalization module.
    """

    def __init__(self, d_hid, eps=1e-3):
        super(LayerNormalization, self).__init__()

        self.eps = eps
        self.a_2 = nn.Parameter(torch.ones(d_hid), requires_grad=True)
        self.b_2 = nn.Parameter(torch.zeros(d_hid), requires_grad=True)

    def forward(self, z):
        if z.size(1) == 1:
            return z

        mu = torch.mean(z, keepdim=True, dim=-1)
        sigma = torch.std(z, keepdim=True, dim=-1)
        ln_out = (z - mu.expand_as(z)) / (sigma.expand_as(z) + self.eps)
        ln_out = ln_out * self.a_2.expand_as(ln_out) + self.b_2.expand_as(ln_out)

        return ln_out


class SupportEncoder(nn.Module):
    """
    Docstring for SupportEncoder
    """

    def __init__(self, d_model, d_inner, dropout=0.1):
        super(SupportEncoder, self).__init__()
        self.proj1 = nn.Linear(d_model, d_inner)
        self.proj2 = nn.Linear(d_inner, d_model)
        self.layer_norm = LayerNormalization(d_model)

        init.xavier_normal_(self.proj1.weight)
        init.xavier_normal_(self.proj2.weight)

        self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU()

    def forward(self, x):
        residual = x
        output = self.relu(self.proj1(x))
        output = self.dropout(self.proj2(output))
        return self.layer_norm(output + residual)


class MatchNet(nn.Module):
    """
    LSTM-based matching network.
    """

    def __init__(self, input_dim, process_steps=4):
        super(MatchNet, self).__init__()
        self.input_dim = input_dim
        self.process_steps = process_steps
        self.match_net = nn.LSTMCell(input_dim, 2 * input_dim)

    def forward(self, support, query):
        assert support.size()[1] == query.size()[1]

        if self.process_steps == 0:
            return F.cosine_similarity(query, support, dim=1)
            # return torch.matmul(query, support.t()).squeeze()

        batch_size = query.size()[0]

        h_r = Variable(torch.zeros(batch_size, 2 * self.input_dim)).cuda()
        c = Variable(torch.zeros(batch_size, 2 * self.input_dim)).cuda()

        for step in range(self.process_steps):
            h_r_, c = self.match_net(query, (h_r, c))
            h = query + h_r_[:, :self.input_dim]
            attn = F.softmax(torch.matmul(h, support.t()), dim=1)
            r = torch.matmul(attn, support)
            h_r = torch.cat((h, r), dim=1)

        matching_scores = torch.matmul(h, support.t()).squeeze()
        return matching_scores



import torch
import torch.nn as nn
import torch.nn.functional as F


class GMUCp_Loss(nn.Module):
    """GMUC+ Loss

    Attributes:
        args: Some pre-set parameters, etc
        model: The UKG model for training.
    """
    def __init__(self, args, model):
        super(GMUCp_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, query_scores, query_scores_var, false_scores, query_confidence, symbolid_ic):
        """
        Args:
            query_scores: The matching scores for link prediction.
            query_scores_var: The prediction value of confidence.
            false_scores: The loss for false set.
            query_confidence: The true value for confidence in query set.
            symbolid_ic: The loss for symbolid-ic value.
        Returns:
            loss: The training loss for back propagation.
        """
        if self.args.num_neg != 1:
            false_scores = false_scores.reshape((query_scores.shape[0], self.args.num_neg))  # resize
            false_scores = torch.mean(false_scores, dim=1)
            false_scores = false_scores.reshape((query_scores.shape[0]))  # resize

        zero_torch = torch.zeros(query_confidence.shape).cuda()
        ones_torch = torch.ones(query_confidence.shape).cuda()
        query_conf_mask = torch.where(query_confidence < self.args.conf_thr, zero_torch, query_confidence)
        # ------ MSE loss -------
        mse_loss = (query_scores_var - query_confidence) ** 2
        mse_loss = self.args.mse_weight * mse_loss.mean()
        # ------ rank loss ------
        rank_loss = self.args.margin - (query_scores - false_scores)
        if self.args.if_conf:
            rank_loss = torch.mean(F.relu(rank_loss) * query_conf_mask)
        else:
            rank_loss = torch.mean(F.relu(rank_loss))
        rank_loss = self.args.rank_weight * rank_loss
        # ic loss
        symbol_ids = symbolid_ic[:, 0].squeeze().long()
        symbol_emb_var_ = self.model.symbol_emb_var(symbol_ids)
        symbol_emb_var_norm = torch.norm(symbol_emb_var_, p=2, dim=1)
        symbol_ics = symbolid_ic[:, 1].squeeze()
        ic_loss = torch.mean(torch.square(self.model.ic_loss_w * symbol_emb_var_norm + self.model.ic_loss_b - symbol_ics))
        ic_loss = self.args.ic_weight * ic_loss

        loss = rank_loss + mse_loss + ic_loss

        return loss



import argparse
import pytorch_lightning as pl
import torch
from collections import defaultdict as ddict
from src.unKR.utils.tools import import_class
import numpy as np


class Config(dict):
    def __getattr__(self, name):
        return self.get(name)

    def __setattr__(self, name, val):
        self[name] = val


class BaseLitModel(pl.LightningModule):
    """
    Generic PyTorch-Lightning class that must be initialized with a PyTorch module.
    """

    def __init__(self, model, args: argparse.Namespace = None, src_list=None, dst_list=None, rel_list=None):
        super().__init__()
        self.train_epoch = 0
        self.model = model
        self.pool = []  # Sample pools
        self.entity = set()  # Entity pools
        self.relation = set()  # Relation pools
        self.semiM = 0  # Maximum number of semi-supervised samples
        self.alpha = 0.02
        self.training_set = []
        self.training_set_triples = []
        self.training_set_triples_list = []
        self.hr2t_train = ddict(set)
        self.rt2h_train = ddict(set)
        self.h2rt_train = ddict(set)
        self.t2rh_train = ddict(set)
        self.args = args
        self.neg_per_positive = args.num_neg  # 负采样数量
        optim_name = args.optim_name
        self.optimizer_class = getattr(torch.optim, optim_name)
        loss_name = args.loss_name
        # self.loss_class = getattr(loss, loss_name)
        self.loss_class = import_class(f"src.unKR.loss.{loss_name}")
        self.loss = self.loss_class(args, model)
        if self.args.model_name == 'SEGNN':
            self.automatic_optimization = False

    @staticmethod
    def add_to_argparse(parser):
        parser.add_argument("--lr", type=float, default=0.1)
        parser.add_argument("--weight_decay", type=float, default=0.01)
        return parser

    def configure_optimizers(self):
        raise NotImplementedError

    def forward(self, x):
        raise NotImplementedError

    def training_step(self, batch, batch_idx):  # pylint: disable=unused-argument
        raise NotImplementedError

    def validation_step(self, batch, batch_idx):  # pylint: disable=unused-argument
        raise NotImplementedError

    def test_step(self, batch, batch_idx):  # pylint: disable=unused-argument
        raise NotImplementedError

    def get_results(self, results, mode):
        """Summarize the results of each batch and calculate the final result of the epoch
        Args:
            results ([type]): The results of each batch
            mode ([type]): Eval or Test
        Returns:
            dict: The final result of the epoch
        """
        conf_list = ["MAE", "MSE"]
        link_list = ["mrr", "mr", "hits@10", "hits@5", "hits@3", "hits@1"]
        link_list_raw = ["raw_mrr", "raw_mr", "raw_hits@10", "raw_hits@5", "raw_hits@3", "raw_hits@1"]

        outputs = ddict(float)
        count_for_conf = np.array([o["count_for_conf"] for o in results]).sum()
        count_for_link = np.array([o["count_for_link"] for o in results]).sum()
        conf_sum = np.array([o["sum_for_conf"].cpu() for o in results]).sum()

        for metric in list(results[0].keys())[1:]:
            if metric in conf_list:
                final_metric = "_".join([mode, metric])
                outputs[final_metric] = np.around(np.array([o[metric] for o in results]).sum() / count_for_conf,
                                                  decimals=6).item()
            elif metric in link_list:
                final_metric = "_".join([mode, metric])
                outputs[final_metric] = np.around(np.array([o[metric] for o in results]).sum() / count_for_link,
                                                  decimals=6).item()
            # elif metric in link_list_high_score_filtered:
            #     final_metric = "_".join([mode, metric])
            #     outputs[final_metric] = np.around(np.array([o[metric] for o in results]).sum() / count_for_link, decimals=3).item()
            elif metric in link_list_raw:
                final_metric = "_".join([mode, metric])
                outputs[final_metric] = np.around(np.array([o[metric] for o in results]).sum() / count_for_link,
                                                  decimals=6).item()
            else:
                outputs["_".join([mode, 'wmr'])] = np.around(
                    np.array([o['wmr'].cpu() for o in results]).sum() / conf_sum, decimals=6).item()
                outputs["_".join([mode, 'wmrr'])] = np.around(
                    np.array([o['wmrr'].cpu() for o in results]).sum() / conf_sum, decimals=6).item()
                outputs["_".join([mode, 'raw_wmr'])] = np.around(
                    np.array([o['raw_wmr'].cpu() for o in results]).sum() / conf_sum, decimals=6).item()
                outputs["_".join([mode, 'raw_wmrr'])] = np.around(
                    np.array([o['raw_wmrr'].cpu() for o in results]).sum() / conf_sum, decimals=6).item()

        return outputs

    def convert_semi_samples_batch(self, neg_samples, neg_scores, pos_samples, mode, device):
        # 计算负样本的总数量
        num_neg_samples = len(pos_samples) * len(neg_samples[0])
        # 复制正样本，以便修改
        semi_samples_formatted = pos_samples.repeat(1, len(neg_samples[0])).view(-1, 4).to(device)
        # 根据mode选择是替换头实体还是尾实体
        if mode == "head-batch":
            neg_samples_formatted = neg_samples.view(-1).long().cpu().numpy()
            semi_samples_formatted[:, 0] = torch.tensor(neg_samples_formatted, dtype=torch.double)
        elif mode == "tail-batch":
            neg_samples_formatted = neg_samples.view(-1).long().cpu().numpy()
            semi_samples_formatted[:, 2] = torch.tensor(neg_samples_formatted, dtype=torch.double)
        else:
            raise ValueError("Invalid mode. Use 'head-batch' or 'tail-batch'.")
        # 将得分扩展成与负样本数量相同的形状
        neg_scores_expanded = neg_scores.view(-1).to(device)
        # 添加得分作为新的第四个元素
        semi_samples_formatted = torch.cat([semi_samples_formatted[:, 0:3], neg_scores_expanded.unsqueeze(1)], dim=1)
        return semi_samples_formatted

    def convert_neg_samples_batch(self, pos_samples, neg_samples, mode):
        # 获取正样本数量和负样本数量
        num_pos_samples = len(pos_samples)
        num_neg_samples_per_pos = len(neg_samples[0])
        num_neg_samples = num_pos_samples * num_neg_samples_per_pos
        # 一次性复制正样本
        pos_samples_formatted = pos_samples.repeat(1, num_neg_samples_per_pos).view(-1, 4)
        # 根据 mode 选择是替换头实体还是尾实体
        if mode == "head-batch" or mode == "head_predict":
            neg_samples_formatted = neg_samples.view(-1).long()
            pos_samples_formatted[:, 0] = neg_samples_formatted.float()
        elif mode == "tail-batch" or mode == "tail_predict":
            neg_samples_formatted = neg_samples.view(-1).long()
            pos_samples_formatted[:, 2] = neg_samples_formatted.float()
        else:
            raise ValueError("Invalid mode. Use 'head-batch', 'tail-batch', 'head_predict', or 'tail_predict'.")
        # 得分设置为0
        pos_samples_formatted[:, 3] = 0.0
        return pos_samples_formatted

    def select_random_elements(self, tensor, n):
        # 获取 tensor 的长度
        num_elements = len(tensor)
        # 生成随机索引
        random_indices = torch.randperm(num_elements)[:n]

        # 根据随机索引选择元素
        selected_elements = tensor[random_indices]

        return selected_elements

    def head_batch(self, h, r, t, neg_size=None):
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_head(t, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]

    def corrupt_head(self, t, r, num_max=1):
        tmp = torch.randint(low=0, high=self.args.num_ent, size=(num_max,)).numpy()
        if not self.args.filter_flag:
            return tmp
        mask = np.in1d(tmp, self.rt2h_train[(r, t)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg

    def tail_batch(self, h, r, t, neg_size=None):
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_tail(h, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]

    def corrupt_tail(self, h, r, num_max=1):
        tmp = torch.randint(low=0, high=self.args.num_ent, size=(num_max,)).numpy()
        if not self.args.filter_flag:
            return tmp
        mask = np.in1d(tmp, self.hr2t_train[(h, r)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg



import torch
import torch.nn as nn


class FocusE_Loss(nn.Module):
    """The loss function of FocusE

    Attributes:
        args: Some pre-set parameters, etc
        model: The FocusE model for training.
    """
    def __init__(self, args, model):
        super(FocusE_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, pos_score, neg_score, pos_sample):
        """Calculating the loss, which includes the negative log-likelihood loss and L3 regularization.

        Args:
            pos_score: Tensor of scores for positive samples.
            neg_score: Tensor of scores for negative samples.
            pos_sample: Tensor of positive samples.

        Returns:
            The total loss of FocusE.
        """
        # neg_score = torch.mean(neg_score, dim=1).unsqueeze(1)
        neg_score = torch.sum(torch.exp(neg_score), dim=1).unsqueeze(1)

        # loss = torch.log(torch.exp(pos_score) / (torch.exp(pos_score) + torch.exp(neg_score)))
        loss = torch.log(torch.exp(pos_score) / (torch.exp(pos_score) + neg_score))
        loss = torch.sum(loss)

        # Use L3 regularization for ComplEx and DistMult
        regularization = self.args.regularization * (
            self.model.ent_emb.weight.norm(p = 3)**3 + \
            self.model.rel_emb.weight.norm(p = 3)**3
        )

        loss = -loss + regularization

        return loss


import torch
import torch.nn.functional as F
import torch.nn as nn


class GTransE_Loss(nn.Module):

    def __init__(self, args, model):
        super(GTransE_Loss, self).__init__()
        self.args = args
        self.model = model
        self.margin = args.margin
        self.alpha = args.alpha

    def forward(self, pos_score, neg_score, pos_sample):
        confidence = pos_sample[:, 3]
        modified_margin = self.margin * (confidence ** self.alpha)
        modified_margin = modified_margin.unsqueeze(1)

        diff = neg_score - pos_score
        loss = F.relu(modified_margin + diff)

        return loss.mean()



import torch
import torch.nn.functional as F
import torch.nn as nn


class UKGE_PSL_Loss(nn.Module):
    """Loss of UKGE with PSL

    """

    def __init__(self, args, model):
        super(UKGE_PSL_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, pos_score, neg_score, PSL_score, pos_sample, neg_sample, PSL_sample):
        confidence = pos_sample[:, 3]  # get confidence of positive triples
        confidence_2 = neg_sample[:, 3]  # get confidence of negative triples
        confidence_3 = PSL_sample[:, 3]  # get confidence of PSL triples

        pos_score = pos_score.squeeze()
        neg_score = neg_score.squeeze()
        PSL_score = PSL_score.squeeze()

        loss_1 = torch.sum((pos_score - confidence) ** 2) # l_pos
        tmp = torch.clamp((confidence_3 - PSL_score), min=0)
        loss_2 = 0.2 * sum(tmp**2) # l_neg
        loss_3 = torch.sum(neg_score ** 2) # l_psl

        loss = loss_1 + loss_2 + loss_3 / neg_score.shape[1]
        return loss



import torch
import torch.nn.functional as F
import torch.nn as nn


class UPGAT_Loss(nn.Module):

    def __init__(self, args, model):
        super(UPGAT_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, pos_score, neg_score, pos_sample, pseudo_score=None, pseudo_sample=None):
        """Calculating the loss score of UPGAT model.

        Args:
            pos_score: The score of positive samples.
            neg_score: The score of negative samples.
            pos_sample: The positive samples.
            pseudo_score: The score of pseudo samples, defaults to None.
            pseudo_sample: The pseudo samples, defaults to None.

        Returns:
            loss: The training loss for back propagation.
        """
        confidence = pos_sample[:, 3]
        pos_score = pos_score.squeeze()
        loss_1 = torch.sum((pos_score - confidence) ** 2)
        loss_2 = torch.sum(neg_score ** 2)
        loss = loss_1 + loss_2 / neg_score.shape[1]
        if pseudo_sample is not None:
            pseudo_confidence = pseudo_sample[:, 3]
            pseudo_score = pseudo_score.squeeze()
            loss_3 = torch.sum((pseudo_score - pseudo_confidence) ** 2)
            loss = loss + loss_3 * self.args.train_bs / self.args.pseudo_bs

        return loss



import numpy as np
from .BaseLitModel import BaseLitModel
from ..eval_task import *
import random


class PASSLEAFLitModel(BaseLitModel):
    """Processing of training, evaluation and testing for PASSLEAF.
    """

    def __init__(self, model, args):
        super().__init__(model, args)

    def forward(self, x):
        return self.model(x)

    @staticmethod
    def add_to_argparse(parser):
        parser.add_argument("--lr", type=float, default=0.1)
        parser.add_argument("--weight_decay", type=float, default=0.01)
        return parser


    def training_step(self, batch, batch_idx):
        """Getting samples and training in KG model.

        Args:
            batch: The training data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            loss: The training loss for back propagation.
        """
        #
        if self.train_epoch >= self.args.T_new_semi:
            # when epoch > T_new_semi, get negative samples Dynamically
            pos_sample = batch["positive_sample"]
            device = pos_sample.device
            data = batch["ori_data"]
            mode = batch["mode"]
            neg_ent_sample = []
            # get semi-supervised samples in the same way as getting negative samples
            if mode == "head-batch":
                for h, r, t, _ in data:
                    neg_head = self.head_batch(h, r, t, self.args.num_neg)
                    neg_ent_sample.append(neg_head)
            else:
                for h, r, t, _ in data:
                    neg_tail = self.tail_batch(h, r, t, self.args.num_neg)
                    neg_ent_sample.append(neg_tail)
            batch['semi_sample'] = torch.LongTensor(np.array(neg_ent_sample)).to(device)

        if self.train_epoch < self.args.T_new_semi:
            # when self.train_epoch < self.args.T_new_semi, train as normal UKGE
            pos_sample = batch["positive_sample"] # get positive samples
            neg_sample = batch["negative_sample"] # get negative samples
            mode = batch["mode"]
            pos_score = self.model(pos_sample) # calculate the score of postive samples
            neg_score = self.model(pos_sample, neg_sample, mode) # calculate the score of negative samples
            head_emb, relation_emb, tail_emb = self.model.tri2emb(pos_sample[:, :3].to(torch.int))
            head_emb = head_emb.squeeze(1)
            relation_emb = relation_emb.squeeze(1)
            tail_emb = tail_emb.squeeze(1)

            regularizer = (torch.sum(torch.square(head_emb)) / head_emb.shape[0]) + \
                          (torch.sum(torch.square(relation_emb)) / relation_emb.shape[0]) + \
                          (torch.sum(torch.square(tail_emb)) / tail_emb.shape[0])

            loss = self.loss(pos_score, neg_score, pos_sample) + 0.005 * regularizer # the same loss as UKGE
            self.log("Train|loss", loss, on_step=False, on_epoch=True)

            if self.train_epoch == 0: # get all entities and relations in epoch 0
                for data in batch['positive_sample']:
                    self.entity.add(data[0])
                    self.entity.add(data[2])
                    self.relation.add(data[1])
                    self.training_set.append(data)

        if self.train_epoch >= self.args.T_new_semi and self.train_epoch < self.args.T_semi_train:
            # begin to add samples to the pool
            pos_sample = batch["positive_sample"]
            neg_sample = batch["negative_sample"]
            semi_sample = batch["semi_sample"] # get semi_samples, the semi_samples were sampled in semi_sample, they will be different in each epoch
            # the number of semi_samples is equal to the number of neg_samples
            mode = batch["mode"]
            device = pos_sample.device
            pos_score = self.model(pos_sample)
            neg_score = self.model(pos_sample, neg_sample, mode)
            semi_score = self.model(pos_sample, semi_sample, mode) # calculate the score of semi-supervised samples to be added to the pool
            semi_sample_format = self.convert_semi_samples_batch(semi_sample, semi_score, pos_sample, mode, device) # convert entities into concrete semi triples
            semi_sample_format_list = torch.tensor(semi_sample_format, device=device, dtype=torch.float64)
            if len(self.pool) <= self.args.max_pool: # add samples to the pool
                self.pool.extend(semi_sample_format_list)

            head_emb, relation_emb, tail_emb = self.model.tri2emb(pos_sample[:, :3].to(torch.int))
            head_emb = head_emb.squeeze(1)
            relation_emb = relation_emb.squeeze(1)
            tail_emb = tail_emb.squeeze(1)

            regularizer = (torch.sum(torch.square(head_emb)) / head_emb.shape[0]) + \
                          (torch.sum(torch.square(relation_emb)) / relation_emb.shape[0]) + \
                          (torch.sum(torch.square(tail_emb)) / tail_emb.shape[0])
            loss = self.loss(pos_score, neg_score, pos_sample) + 0.005 * regularizer
            self.log("Train|loss", loss, on_step=False, on_epoch=True)

        if self.train_epoch >= self.args.T_semi_train:
            # begin to use samples in the pool
            n_generated_samples = len(batch['positive_sample']) * self.neg_per_positive  # the number of all samples(neg + semi) used for training
            n_semi_samples = int(
                min(n_generated_samples * 0.8, max(0,
                                                   -self.args.T_semi_train + self.train_epoch + 1) * n_generated_samples * self.args.alpha_PASSLEAF)) # number of semi samples to be used in this epoch
            N_neg = n_generated_samples - n_semi_samples # number of negative samples to be used in this epoch

            pos_sample = batch["positive_sample"]
            neg_sample = batch["negative_sample"]
            mode = batch["mode"]

            semi_sample = random.sample(self.pool, n_semi_samples)  # selected semi_samples from the pool
            semi_sample_stacked = torch.stack(semi_sample)

            format_neg_sample = self.convert_neg_samples_batch(pos_sample, neg_sample, mode) # convert entities into concrete neg triples
            selected_neg_samples = self.select_random_elements(format_neg_sample, N_neg) # selected neg_samples in the epoch
            pos_score = self.model(pos_sample) # get the score of positive triples
            semi_score = self.model(semi_sample_stacked)
            # the format of pos_sample, semi_sample_stacked, selected_neg_samples are the same
            # So we can directly emulate the calculation method of pos_score to calculate neg_score and semi_score
            neg_score = self.model(selected_neg_samples)
            head_emb, relation_emb, tail_emb = self.model.tri2emb(pos_sample[:, :3].to(torch.int))
            head_emb = head_emb.squeeze(1)
            relation_emb = relation_emb.squeeze(1)
            tail_emb = tail_emb.squeeze(1)

            regularizer = (torch.sum(torch.square(head_emb)) / head_emb.shape[0]) + \
                          (torch.sum(torch.square(relation_emb)) / relation_emb.shape[0]) + \
                          (torch.sum(torch.square(tail_emb)) / tail_emb.shape[0])
            # The loss will change when add semi_score
            loss = self.loss(pos_score, neg_score, pos_sample, semi_score, semi_sample_stacked) + 0.005 * regularizer
            self.log("Train|loss", loss, on_step=False, on_epoch=True)

            pos_sample = batch["positive_sample"]
            neg_sample = batch["negative_sample"]
            semi_sample = batch["semi_sample"]

            mode = batch["mode"]
            device = pos_sample.device
            semi_score = self.model(pos_sample, semi_sample, mode)
            semi_sample_format = self.convert_semi_samples_batch(semi_sample, semi_score, pos_sample, mode, device) # get semi-samples in the same way as negative samples
            semi_sample_format_list = torch.tensor(semi_sample_format, device=device, dtype=torch.float64)
            if len(self.pool) <= self.args.max_pool: # add samples to the pool
                self.pool.extend(semi_sample_format_list)

        return loss

    def training_epoch_end(self, loss) -> None:
        # get hr2t_train and rt2h_train when epoch is 0, they will be used in the sampling of semi samples
        if self.train_epoch == 0:
            for i in self.training_set:
                self.training_set_triples_list.append(i[0:3].tolist())
            for h, r, t in self.training_set_triples_list:
                self.hr2t_train[(h, r)].add(t)
                self.rt2h_train[(r, t)].add(h)
        self.train_epoch += 1


    def validation_step(self, batch, batch_idx):
        """Getting samples and validation in KG model.

        Args:
            batch: The evalutaion data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            results: MAE, MSE, mrr, mr, wmrr, wmr, hits@1,3,5,10, raw_mrr, raw_wmr, raw_wmrr, raw_wmr, raw_hits@1,3,5,10
        """
        results = dict()

        results["count_for_conf"] = batch['positive_sample'].shape[0]
        MAE, MSE = conf_predict(batch, self.model)
        results["MAE"] = MAE.item()
        results["MSE"] = MSE.item()

        prediction = "tail"
        confidence = self.args.confidence_filter
        ranks = link_predict(batch, self.model, prediction=prediction)
        ranks_link_predict = link_predict_filter(batch, self.model, confidence, prediction=prediction) # get the ranks of entities in link prediction
        ranks_link_predict_raw = link_predict_raw(batch, self.model, confidence, prediction=prediction) # get the raw rank (no filter when calculate the rank)
        results["count_for_link"] = torch.numel(ranks_link_predict)
        results["mrr"] = torch.sum(1.0 / ranks_link_predict).item()
        results["mr"] = torch.sum(ranks_link_predict).item()
        results["raw_count_for_link"] = torch.numel(ranks_link_predict_raw)
        results["raw_mrr"] = torch.sum(1.0 / ranks_link_predict_raw).item()
        results["raw_mr"] = torch.sum(ranks_link_predict_raw).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks_link_predict[ranks_link_predict <= k])
            results['raw_hits@{}'.format(k)] = torch.numel(ranks_link_predict_raw[ranks_link_predict_raw <= k])

        """calculate WMR   Weighted-MR """
        pos_triple = batch["positive_sample"]
        mask = pos_triple[:, -1] >= confidence
        if prediction == "all":
            conf = torch.cat([batch['positive_sample'][:, 3]] * 2)
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        else:
            conf = batch['positive_sample'][:, 3]
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        return results


    def validation_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Eval")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        """Getting samples and validation in KG model.

        Args:
            batch: The evalutaion data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            results: MAE, MSE, mrr, mr, wmrr, wmr, hits@1,3,5,10, raw_mrr, raw_wmr, raw_wmrr, raw_wmr, raw_hits@1,3,5,10
        """
        results = dict()

        results["count_for_conf"] = batch['positive_sample'].shape[0]
        MAE, MSE = conf_predict(batch, self.model)
        results["MAE"] = MAE.item()
        results["MSE"] = MSE.item()

        prediction = "tail"
        confidence = self.args.confidence_filter
        ranks = link_predict(batch, self.model, prediction=prediction)
        ranks_link_predict = link_predict_filter(batch, self.model, confidence, prediction=prediction)
        ranks_link_predict_raw = link_predict_raw(batch, self.model, confidence, prediction=prediction)
        results["count_for_link"] = torch.numel(ranks_link_predict)
        results["mrr"] = torch.sum(1.0 / ranks_link_predict).item()
        results["mr"] = torch.sum(ranks_link_predict).item()
        results["raw_count_for_link"] = torch.numel(ranks_link_predict_raw)
        results["raw_mrr"] = torch.sum(1.0 / ranks_link_predict_raw).item()
        results["raw_mr"] = torch.sum(ranks_link_predict_raw).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks_link_predict[ranks_link_predict <= k])
            results['raw_hits@{}'.format(k)] = torch.numel(ranks_link_predict_raw[ranks_link_predict_raw <= k])

        """calculate WMR   Weighted-MR """
        pos_triple = batch["positive_sample"]
        mask = pos_triple[:, -1] >= confidence
        if prediction == "all":
            conf = torch.cat([batch['positive_sample'][:, 3]] * 2)
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        else:
            conf = batch['positive_sample'][:, 3]
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        return results

    def test_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Test")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def configure_optimizers(self):
        """Setting optimizer and lr_scheduler.
        Returns:
            optim_dict: Record the optimizer and lr_scheduler, type: dict.
        """
        milestones = int(self.args.max_epochs / 8)
        beta1 = 0.9  # set β1
        beta2 = 0.99  # set β2
        optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr, betas=(beta1, beta2))
        StepLR = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[milestones], gamma=0.1)
        optim_dict = {'optimizer': optimizer, 'lr_scheduler': StepLR}
        return optim_dict



import torch.nn as nn
import torch
from .model import Model


class UKGE(Model):
    """`Embedding Uncertain Knowledge Graphs`_ (UKGE), which represents the relationships as translations in the embedding space for uncertain knowledge graph.

    Attributes:
        args: Model configuration parameters.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation embedding, shape:[num_rel, emb_dim].
        w: Weight of mapping function.
        b: Bias of mapping function.

    .. _Embedding Uncertain Knowledge Graphs: https://ojs.aaai.org/index.php/AAAI/article/view/4210
    """

    def __init__(self, args):
        super(UKGE, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.w = torch.nn.Parameter(torch.tensor(1.0), requires_grad=True)
        self.b = torch.nn.Parameter(torch.tensor(0.0), requires_grad=True)

        self.init_emb()

    def init_emb(self):
        """
            Initialize the entity and relation embeddings in the form of a uniform distribution.
        """

        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        nn.init.xavier_uniform_(self.ent_emb.weight.data)
        nn.init.xavier_uniform_(self.rel_emb.weight.data)

    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.

        The formula for calculating the score is :math:`h^{\top} \operatorname{diag}(r) t`

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        if mode == 'head-batch':
            score = head_emb * (relation_emb * tail_emb)
        else:
            score = (head_emb * relation_emb) * tail_emb

        score = score.sum(dim=-1)
        score = score.to(self.args.gpu)

        """ 1.Bounded rectifier """
        # shape = score.shape
        # tmp_max = torch.max(self.w * score + self.b, torch.zeros(shape, device=self.args.gpu))
        # score = torch.min(tmp_max, torch.ones(shape, device=self.args.gpu))

        """ 2.Logistic function"""
        score = torch.sigmoid(self.w * score + self.b)

        return score

    def forward(self, triples, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t, c), shape:[batch_size, 4].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        triples = triples[:, :3].to(torch.int)
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        triples = batch["positive_sample"]
        triples = triples[:, :3].to(torch.int)
        triples = triples.to(self.args.gpu)

        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
from torch.autograd import Variable


class Matcher(nn.Module):
    """
    Gaussian Matching Networks.
    """

    def __init__(self, args, num_symbols, embed=None):
        super(Matcher, self).__init__()
        self.args = args
        self.pad_idx = num_symbols
        self.symbol_emb = nn.Embedding(num_symbols + 1, self.args.emb_dim, padding_idx=num_symbols)
        self.symbol_var_emb = nn.Embedding(num_symbols + 1, self.args.emb_dim, padding_idx=num_symbols)
        self.num_symbols = num_symbols
        self.layer_norm = LayerNormalization(2 * self.args.emb_dim)
        dropout = self.args.dropout

        if self.args.random_embed:
            self.args.use_pretrain = False
        else:
            self.args.use_pretrain = True

        self.dropout = nn.Dropout(dropout)

        # aggregation encoder and decoder
        self.set_rnn_encoder = nn.LSTM(2 * self.args.emb_dim, 2 * self.args.emb_dim, 1, bidirectional=False)
        self.set_rnn_decoder = nn.LSTM(2 * self.args.emb_dim, 2 * self.args.emb_dim, 1, bidirectional=False)

        self.set_FC_encoder = nn.Linear(3 * 2 * self.args.emb_dim, 2 * self.args.emb_dim)
        self.set_FC_decoder = nn.Linear(2 * self.args.emb_dim, 3 * 2 * self.args.emb_dim)

        # neighbor encoder attention
        self.neigh_att_W = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.neigh_att_u = nn.Linear(self.args.emb_dim, 1)
        self.neigh_var_att_W = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.neigh_var_att_u = nn.Linear(self.args.emb_dim, 1)
        # aggregation attention
        self.set_att_W = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.set_att_u = nn.Linear(self.args.emb_dim, 1)
        self.set_var_att_W = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.set_var_att_u = nn.Linear(self.args.emb_dim, 1)

        self.bn = nn.BatchNorm1d(2 * self.args.emb_dim)
        self.softmax = nn.Softmax(dim=1)

        self.FC_query_g = nn.Linear(2 * self.args.emb_dim, 2 * self.args.emb_dim)
        self.FC_support_g_encoder = nn.Linear(2 * self.args.emb_dim, 2 * self.args.emb_dim)

        init.xavier_normal_(self.neigh_att_W.weight)
        init.xavier_normal_(self.neigh_att_u.weight)
        init.xavier_normal_(self.set_att_W.weight)
        init.xavier_normal_(self.set_att_u.weight)
        init.xavier_normal_(self.FC_query_g.weight)
        init.xavier_normal_(self.FC_support_g_encoder.weight)

        # Here is symbol to embedding
        if self.args.use_pretrain:
            self.symbol_emb.weight.data.copy_(torch.from_numpy(embed))
            self.symbol_var_emb.weight.data.copy_(torch.from_numpy(embed))

        d_model = self.args.emb_dim * 2
        self.support_encoder = SupportEncoder(d_model, 2 * d_model, dropout)
        self.query_encoder = QueryEncoder(d_model, self.args.process_steps)

        self.w = nn.Parameter(torch.FloatTensor(1), requires_grad=True)
        self.b = nn.Parameter(torch.FloatTensor(1), requires_grad=True)

    def neighbor_encoder(self, connections, num_neighbors):

        num_neighbors = num_neighbors.unsqueeze(1)
        relations = connections[:, :, 0].squeeze(-1).long()  # neighbor relation
        entities = connections[:, :, 1].squeeze(-1).long()  # neighor entity
        confidences = connections[:, :, 2]  # get neighor confidence (batch, neighbor)

        (batch_size, neighbor_size) = confidences.shape
        confidences = confidences.reshape((batch_size, neighbor_size, 1)).repeat((1, 1, 2 * self.args.emb_dim))

        rel_embeds = self.dropout(self.symbol_emb(relations))  # (batch, neighbor_num, emb_dim)
        ent_embeds = self.dropout(self.symbol_emb(entities))  # (batch, neighbor_num, emb_dim)
        concat_embeds = torch.cat((rel_embeds, ent_embeds),
                                  dim=-1)  # (batch, neighbor_num, 2*emb_dim)

        rel_var_embeds = self.dropout(self.symbol_var_emb(relations))  # (batch, neighbor_num, emb_dim)
        ent_var_embeds = self.dropout(self.symbol_var_emb(entities))  # (batch, neighbor_num, emb_dim)
        concat_var_embeds = torch.cat((rel_var_embeds, ent_var_embeds),
                                      dim=-1)  # (batch, neighbor_num, 2*emb_dim)

        # position embedding
        out = self.neigh_att_W(concat_embeds).tanh()
        att_w = self.neigh_att_u(out)
        att_w = self.softmax(att_w).view(concat_embeds.size()[0], 1, 30)
        out = torch.bmm(att_w, ent_embeds).view(concat_embeds.size()[0], self.args.emb_dim)

        # variance embedding
        out_var = self.neigh_var_att_W(concat_var_embeds).tanh()
        att_var_w = self.neigh_var_att_u(out_var)
        att_var_w = self.softmax(att_var_w).view(concat_var_embeds.size()[0], 1, 30)
        out_var = torch.bmm(att_var_w, ent_embeds).view(concat_var_embeds.size()[0], self.args.emb_dim)

        return out.tanh(), out_var.tanh()

    def score_func(self, support, support_meta, query, query_meta):
        raise NotImplementedError

    def forward(self, support, support_meta, query, query_meta, false, false_meta):
        raise NotImplementedError


class LayerNormalization(nn.Module):
    """
    Layer normalization module.
    """
    def __init__(self, d_hid, eps=1e-3):
        super(LayerNormalization, self).__init__()

        self.eps = eps
        self.a_2 = nn.Parameter(torch.ones(d_hid), requires_grad=True)
        self.b_2 = nn.Parameter(torch.zeros(d_hid), requires_grad=True)

    def forward(self, z):
        if z.size(1) == 1:
            return z

        mu = torch.mean(z, keepdim=True, dim=-1)
        sigma = torch.std(z, keepdim=True, dim=-1)
        ln_out = (z - mu.expand_as(z)) / (sigma.expand_as(z) + self.eps)
        ln_out = ln_out * self.a_2.expand_as(ln_out) + self.b_2.expand_as(ln_out)

        return ln_out


class SupportEncoder(nn.Module):
    """
    The encoder for support set.
    """
    def __init__(self, d_model, d_inner, dropout=0.1):
        super(SupportEncoder, self).__init__()
        self.proj1 = nn.Linear(d_model, d_inner)
        self.proj2 = nn.Linear(d_inner, d_model)
        self.layer_norm = LayerNormalization(d_model)

        init.xavier_normal_(self.proj1.weight)
        init.xavier_normal_(self.proj2.weight)

        self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU()

    def forward(self, x):
        residual = x
        output = self.relu(self.proj1(x))
        output = self.dropout(self.proj2(output))
        return self.layer_norm(output + residual)


class QueryEncoder(nn.Module):
    """
    The encoder for query set, similar to
    https://proceedings.neurips.cc/paper_files/paper/2016/file/90e1357833654983612fb05e3ec9148c-Paper.pdf.
    """
    def __init__(self, input_dim, process_step=4):
        super(QueryEncoder, self).__init__()
        self.input_dim = input_dim
        self.process_step = process_step
        self.process = nn.LSTMCell(input_dim, 2 * input_dim)

    def forward(self, support, query):
        assert support.size()[1] == query.size()[1]

        if self.process_step == 0:
            return query

        batch_size = query.size()[0]
        h_r = Variable(torch.zeros(batch_size, 2 * self.input_dim)).cuda()
        c = Variable(torch.zeros(batch_size, 2 * self.input_dim)).cuda()

        for step in range(self.process_step):
            h_r_, c = self.process(query, (h_r, c))
            h = query + h_r_[:, :self.input_dim]  # (batch_size, query_dim)
            attn = F.softmax(torch.matmul(h, support.t()), dim=1)
            r = torch.matmul(attn, support)  # (batch_size, support_dim)
            h_r = torch.cat((h, r), dim=1)

        return h


class GMUC(Matcher):

    """`Gaussian Metric Learning for Few-Shot Uncertain Knowledge Graph Completion`_ (GMUC),
        which proposes a novel method to complete few-shot UKGs based on Gaussian metric learning.

    Attributes:
        args: Model configuration parameters.
        num_symbols: The sum of the number of entities and relationships.
        embed: Pretrained embedding.

    .. _Gaussian Metric Learning for Few-Shot Uncertain Knowledge Graph Completion:
        https://link.springer.com/chapter/10.1007/978-3-030-73194-6_18
    """

    def __init__(self, arg, num_symbols, embed=None):
        super(GMUC, self).__init__(arg, num_symbols, embed)
        d_model = self.args.emb_dim * 2

        self.matchnet_mean = MatchNet(d_model, self.args.process_steps)
        self.matchnet_var = MatchNet(d_model, self.args.process_steps)
        if 'N3' in self.args.data_path:
            self.matchnet_mean = MatchNet(d_model, self.args.process_steps)
            self.matchnet_var = MatchNet(d_model, 1)
        self.gcn_w = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)
        self.gcn_w_var = nn.Linear(2 * self.args.emb_dim, self.args.emb_dim)

    def score_func(self, support, support_meta, query, query_meta):
        """Calculating the score of query triples with support set.

        Args:
            support: Support set.
            support_meta: Neighbor information of the head and tail entities in support set.
            query: Query set.
            query_meta: Neighbor information of the head and tail entities in query set.

        Returns:
            matching_scores: Matching scores for link prediction.
            matching_scores_var: The prediction of confidence.
            ae_loss: The loss of encoder input and decoder output.
        """
        query_left_connections, query_left_degrees, query_right_connections, query_right_degrees = query_meta
        support_left_connections, support_left_degrees, support_right_connections, support_right_degrees = support_meta

        # ======================== 1. Guassin Neighbor Encoder ========================
        query_left, query_var_left = self.neighbor_encoder(query_left_connections, query_left_degrees)
        query_right, query_var_right = self.neighbor_encoder(query_right_connections, query_right_degrees)
        support_left, support_var_left = self.neighbor_encoder(support_left_connections, support_left_degrees)
        support_right, support_var_right = self.neighbor_encoder(support_right_connections, support_right_degrees)
        support_confidence = support[:, 2]  # get confidence of support set (support_num)
        # ----------------- mean encoder ------------------
        query_neighbor = torch.cat((query_left, query_right), dim=-1)  # tanh
        support_neighbor = torch.cat((support_left, support_right), dim=-1)  # tanh
        support = support_neighbor
        query = query_neighbor
        support_g = self.support_encoder(support)  # (3, 200)
        query_g = self.support_encoder(query)  # (batch, 2 * embedd_dim)

        # position encoder and decoder
        support_g_0 = support_g.view(3, 1, 2 * self.args.emb_dim)  # input of encoder
        support_g_encoder, support_g_state = self.set_rnn_encoder(support_g_0)

        # decoder
        support_g_decoder = support_g_encoder[-1].view(1, -1, 2 * self.args.emb_dim)
        support_g_decoder_state = support_g_state
        decoder_set = []
        for idx in range(3):
            support_g_decoder, support_g_decoder_state = self.set_rnn_decoder(support_g_decoder,
                                                                              support_g_decoder_state)
            decoder_set.append(support_g_decoder)
        decoder_set = torch.cat(decoder_set, dim=0)  # output of decoder
        ae_loss = nn.MSELoss()(support_g_0, decoder_set.detach())  # calculate loss of encoder input and decoder output
        # encoder
        support_g_encoder = support_g_encoder.view(3, 2 * self.args.emb_dim)
        support_g_encoder = support_g_0.view(3, 2 * self.args.emb_dim) + support_g_encoder  # output of encoder

        # add attention
        support_g_att = self.set_att_W(support_g_encoder).tanh()
        att_w = self.set_att_u(support_g_att)
        att_w = self.softmax(att_w)
        support_g_encoder = torch.matmul(support_g_encoder.transpose(0, 1), att_w)
        support_g_encoder = support_g_encoder.transpose(0, 1)
        support_g_encoder = support_g_encoder.view(1, 2 * self.args.emb_dim)

        support = support_g_encoder
        query = query_g

        # ----------------- variance encoder ------------------
        query_var_neighbor = torch.cat((query_var_left, query_var_right), dim=-1)  # tanh
        support_var_neighbor = torch.cat((support_var_left, support_var_right), dim=-1)  # tanh
        support_var = support_var_neighbor
        query_var = query_var_neighbor
        support_var = torch.mean(support_var_neighbor, dim=0, keepdim=True)

        #  ======================== 2. Guassin Matching Networks ========================
        # ----------------- mean match ------------------
        matching_scores = self.matchnet_mean(support, None, query, None)
        # ----------------- varaince match ------------------
        matching_scores_var = self.matchnet_var(support_var, None, query_var, None)
        matching_scores_var = torch.sigmoid(self.w + matching_scores_var + self.b)

        return matching_scores, matching_scores_var, ae_loss

    def forward(self, support, support_meta, query, query_meta, false, false_meta):
        """The functions used in the training phase.

        Args:
            support: Support set.
            support_meta: Neighbor information of the head and tail entities in support set.
            query: Query set.
            query_meta: Neighbor information of the head and tail entities in query set.
            false: False set.
            false_meta: Neighbor information of the head and tail entities in false set.

        Returns:
            query_scores: The match score between the query set and the support set.
            query_scores_var: Support set confidence prediction values.
            query_ae_loss: The loss of encoder input and decoder output in query set.
            false_scores: The match score between the false set and the support set.
            false_scores_var: False set confidence prediction values.
            false_ae_loss: The loss of encoder input and decoder output in false set.
            query_confidence: The confidence for query set.
        """
        query_scores, query_scores_var, query_ae_loss = self.score_func(support, support_meta, query, query_meta)
        false_scores, false_scores_var, false_ae_loss = self.score_func(support, support_meta, false, false_meta)
        query_confidence = query[:, 2]

        return query_scores, query_scores_var, query_ae_loss, false_scores, false_scores_var, false_ae_loss, query_confidence


class MatchNet(nn.Module):
    """
    RNN Match networks for support set and query set, similar to
    https://proceedings.neurips.cc/paper_files/paper/2016/file/90e1357833654983612fb05e3ec9148c-Paper.pdf.
    """
    def __init__(self, input_dim, process_step=4):
        super(MatchNet, self).__init__()
        self.input_dim = input_dim
        self.process_step = process_step
        self.process = nn.LSTMCell(input_dim, 2 * input_dim)

    def forward(self, support_mean, support_var, query_mean, query_var):
        assert support_mean.size()[1] == query_mean.size()[1]

        if self.process == 1:
            return torch.matmul(query_mean, support_mean.t()).squeeze()

        batch_size = query_mean.size()[0]
        h_r = Variable(torch.zeros(batch_size, 2 * self.input_dim)).cuda()
        c = Variable(torch.zeros(batch_size, 2 * self.input_dim)).cuda()

        for step in range(self.process_step):
            h_r_, c = self.process(query_mean, (h_r, c))
            h = query_mean + h_r_[:, :self.input_dim]  # (batch_size, query_dim)

            attn = F.softmax(torch.matmul(h, support_mean.t()), dim=1)

            r = torch.matmul(attn, support_mean)  # (batch_size, support_dim)
            h_r = torch.cat((h, r), dim=1)

        matching_scores = torch.matmul(h, support_mean.t()).squeeze()
        return matching_scores



from .DataPreprocess import *
from torch.autograd import Variable


class UKGEUniSampler(UKGEBaseSampler):
    """Random negative sampling 
    Filtering out positive samples and selecting some samples randomly as negative samples.

    Attributes:
        cross_sampling_flag: The flag of cross sampling head and tail negative samples.
    """

    def __init__(self, args):
        super().__init__(args)
        self.cross_sampling_flag = 0

    def sampling(self, data):
        """Filtering out positive samples and selecting some samples randomly as negative samples.

        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """

        batch_data = {}
        neg_ent_sample = []
        subsampling_weight = []
        self.cross_sampling_flag = 1 - self.cross_sampling_flag
        if self.cross_sampling_flag == 0:
            batch_data['mode'] = "head-batch"
            for h, r, t, _ in data:
                neg_head = self.head_batch(h, r, t, self.args.num_neg)
                neg_ent_sample.append(neg_head)
                if self.args.use_weight:
                    weight = self.count[(h, r)] + self.count[(t, -r - 1)]
                    subsampling_weight.append(weight)
        else:
            batch_data['mode'] = "tail-batch"
            for h, r, t, _ in data:
                neg_tail = self.tail_batch(h, r, t, self.args.num_neg)
                neg_ent_sample.append(neg_tail)
                if self.args.use_weight:
                    weight = self.count[(h, r)] + self.count[(t, -r - 1)]
                    subsampling_weight.append(weight)
        batch_data["ori_data"] = data # storage the original data, it will be used in PASSLEAFLitModel.
        batch_data["positive_sample"] = torch.tensor(np.array(data))
        batch_data['negative_sample'] = torch.LongTensor(np.array(neg_ent_sample))
        batch_data['true_head'] = self.rt2h_train
        batch_data['true_tail'] = self.hr2t_train
        if self.args.use_weight:
            batch_data["subsampling_weight"] = torch.sqrt(1 / torch.tensor(subsampling_weight))

        # When train student_model in UPGAT, get pseudo_sample and adj_matrix as below.
        if self.args.model_name == 'UPGAT':
            train_data = torch.tensor([item[:3] for item in self.train_triples])
            if not self.args.teacher_model:
                pseudo_data = torch.tensor([item[:3] for item in self.pseudo_triples])
                train_data = torch.cat((train_data, pseudo_data), dim=0)
                try:
                    batch_data["pseudo_sample"] = torch.stack(next(self.pseudo_dataiter), dim=1)
                except StopIteration:
                    pseudo_dataloader = DataLoader(self.pseudo_triples, batch_size=self.args.pseudo_bs,
                                                   num_workers=self.args.num_workers, pin_memory=True, shuffle=True,
                                                   drop_last=True)
                    self.pseudo_dataiter = iter(pseudo_dataloader)
                    batch_data["pseudo_sample"] = torch.stack(next(self.pseudo_dataiter),dim=1)
            head, rela, tail = train_data.t()
            self.adj_matrix = (torch.stack((head, tail)), rela)
            batch_data["adj_matrix"] = self.adj_matrix
        return batch_data

    def uni_sampling(self, data):
        batch_data = {}
        neg_head_list = []
        neg_tail_list = []
        for h, r, t, _ in data:
            neg_head = self.head_batch(h, r, t, self.args.num_neg)
            neg_head_list.append(neg_head)
            neg_tail = self.tail_batch(h, r, t, self.args.num_neg)
            neg_tail_list.append(neg_tail)

        batch_data["positive_sample"] = torch.LongTensor(np.array(data))
        batch_data['negative_head'] = torch.LongTensor(np.arrary(neg_head_list))
        batch_data['negative_tail'] = torch.LongTensor(np.arrary(neg_tail_list))
        return batch_data

    def get_sampling_keys(self):
        if self.args.model_name == 'UPGAT':
            if self.args.teacher_model:
                return ['positive_sample', 'negative_sample', 'mode', 'adj_matrix']
            else:
                return ['positive_sample', 'negative_sample', 'mode', 'adj_matrix', 'pseudo_sample']
        else:
            return ['positive_sample', 'negative_sample', 'mode']


class UKGEPSLSampler(UKGEBaseSampler):
    """Random negative sampling Filtering out positive samples and selecting some samples randomly as negative samples.
    UKGEPSLSampler is for UKGE_PSL

    Attributes:
        cross_sampling_flag: The flag of cross sampling head and tail negative samples.
    """

    def __init__(self, args):
        super().__init__(args)
        self.cross_sampling_flag = 0

    def sampling(self, data):
        """Filtering out positive samples and selecting some samples randomly as negative samples.

        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The training data.
        """

        batch_data = {}
        neg_ent_sample = []  # negative samples
        PSL_sample = []  # PSL samples
        subsampling_weight = []
        self.cross_sampling_flag = 1 - self.cross_sampling_flag
        if self.cross_sampling_flag == 0:
            batch_data['mode'] = "head-batch"
            for h, r, t, _ in data:
                neg_head = self.head_batch(h, r, t, self.args.num_neg)
                neg_ent_sample.append(neg_head)
                if self.args.use_weight:
                    weight = self.count[(h, r)] + self.count[(t, -r - 1)]
                    subsampling_weight.append(weight)
        else:
            batch_data['mode'] = "tail-batch"
            for h, r, t, _ in data:
                neg_tail = self.tail_batch(h, r, t, self.args.num_neg)
                neg_ent_sample.append(neg_tail)
                if self.args.use_weight:
                    weight = self.count[(h, r)] + self.count[(t, -r - 1)]
                    subsampling_weight.append(weight)

        PSL_ent_num = int(self.RatioOfPSL * len(data)) # calculate the number of PSL samples in each batch, storage the PSL samples in batch_data['PSL_sample']

        if len(self.PSL_triples) >= PSL_ent_num + 1:
            for i in range(PSL_ent_num + 1):
                PSL_sample.append(self.PSL_triples[i])
            self.PSL_triples = self.PSL_triples[PSL_ent_num + 1:]
        else:
            for i in range(len(self.PSL_triples)):
                PSL_sample.append(self.PSL_triples[i])
        batch_data["positive_sample"] = torch.tensor(np.array(data))
        batch_data["negative_sample"] = torch.tensor(np.array(neg_ent_sample))
        batch_data['PSL_sample'] = torch.tensor(np.array(PSL_sample))

        batch_data['true_head'] = self.h2rt_train
        batch_data['true_tail'] = self.hr2t_train
        if self.args.use_weight:
            batch_data["subsampling_weight"] = torch.sqrt(1 / torch.tensor(subsampling_weight))

        return batch_data

    def uni_sampling(self, data):
        batch_data = {}
        neg_head_list = []
        neg_tail_list = []
        for h, r, t, _ in data:
            neg_head = self.head_batch(h, r, t, self.args.num_neg)
            neg_head_list.append(neg_head)
            neg_tail = self.tail_batch(h, r, t, self.args.num_neg)
            neg_tail_list.append(neg_tail)

        batch_data["positive_sample"] = torch.LongTensor(np.array(data))
        batch_data['negative_head'] = torch.LongTensor(np.arrary(neg_head_list))
        batch_data['negative_tail'] = torch.LongTensor(np.arrary(neg_tail_list))
        return batch_data

    def get_sampling_keys(self):
        return ['positive_sample', 'negative_sample', 'mode']

class UKGETestSampler(object):
    """Sampling triples and recording positive triples for testing.
    We offer two test sample methods: one is to use the same method as neuralkg to process the test set,
    and the other is to only use high confidence samples from the test set for testing

    Attributes:
        sampler: The function of training sampler.
        hr2t_all: Record the tail corresponding to the same head and relation.
        rt2h_all: Record the head corresponding to the same tail and relation.
        hr2t_all_high_score: Record the tail corresponding to the same head and relation (only for high score samples in test).
        rt2h_all_high_score: Record the head corresponding to the same tail and relation (only for high score samples in test).
        num_ent: The count of entities.
    """

    def __init__(self, sampler):
        self.sampler = sampler
        self.hr2t_all = ddict(set)
        self.rt2h_all = ddict(set)
        self.hr2t_all_high_score = ddict(set)
        self.rt2h_all_high_score = ddict(set)
        self.get_hr2t_rt2h_from_all()
        self.num_ent = sampler.args.num_ent

    def get_hr2t_rt2h_from_all(self):
        """Get the set of hr2t and rt2h from all datasets(train, valid, and test), the data type is tensor.
        Update:
            self.hr2t_all: The set of hr2t.
            self.rt2h_all: The set of rt2h.
            self.hr2t_all_high_score: The set of hr2t (only for high score samples in test).
            self.rt2h_all_high_score: The set of rt2h (only for high score samples in test).
        """
        self.all_true_triples = self.sampler.get_all_true_triples()
        for h, r, t, s in self.all_true_triples:
            self.hr2t_all[(h, r)].add(t)
            self.rt2h_all[(r, t)].add(h)
            if s >= self.sampler.args.confidence_filter:
                self.hr2t_all_high_score[(h, r)].add(t)
                self.rt2h_all_high_score[(r, t)].add(h)
        for h, r in self.hr2t_all:
            self.hr2t_all[(h, r)] = torch.tensor(list(self.hr2t_all[(h, r)]))
        for h, r in self.hr2t_all_high_score:
            self.hr2t_all_high_score[(h, r)] = torch.tensor(list(self.hr2t_all_high_score[(h, r)]))
        for r, t in self.rt2h_all:
            self.rt2h_all[(r, t)] = torch.tensor(list(self.rt2h_all[(r, t)]))
        for r, t in self.rt2h_all_high_score:
            self.rt2h_all_high_score[(r, t)] = torch.tensor(list(self.rt2h_all_high_score[(r, t)]))

    def sampling(self, data):
        """Sampling triples and recording positive triples for testing.

        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The data used to be evaluated.
        """
        batch_data = {}
        head_label = torch.zeros(len(data), self.num_ent)
        tail_label = torch.zeros(len(data), self.num_ent)
        head_label_filter = torch.zeros(len(data), self.num_ent)
        tail_label_filter = torch.zeros(len(data), self.num_ent)
        for idx, triple in enumerate(data):
            head, rel, tail, score = triple
            head_label[idx][self.rt2h_all[(rel, tail)]] = 1.0
            tail_label[idx][self.hr2t_all[(head, rel)]] = 1.0
            if self.rt2h_all_high_score[(rel, tail)] != set():
                head_label_filter[idx][self.rt2h_all_high_score[(rel, tail)]] = 1.0
            if self.hr2t_all_high_score[(head, rel)] != set():
                tail_label_filter[idx][self.hr2t_all_high_score[(head, rel)]] = 1.0

        batch_data["positive_sample"] = torch.tensor(data)
        batch_data["head_label"] = head_label
        batch_data["tail_label"] = tail_label
        batch_data["head_label_filter"] = head_label_filter
        batch_data["tail_label_filter"] = tail_label_filter
        return batch_data

    def construct_hr_map(self, data):
        hr_map = {}
        for h, r, t, w in data:
            w = float(w)

            if hr_map.get(h) == None:
                hr_map[h] = {}
            if hr_map[h].get(r) == None:
                hr_map[h][r] = {t: w}
            else:
                hr_map[h][r][t] = w

        return hr_map

    def get_sampling_keys(self):
        return ["positive_sample", "head_label", "tail_label"]
        # return ["positive_sample"]


class GMUCSampler(GMUCBaseSampler):
    """GMUC sampling
        Process task-based data.
    """

    def __init__(self, args):
        super().__init__(args)

    def sampling(self, data):
        """Filtering out positive samples and selecting some samples randomly as negative samples.

        Args:
            data: A task/relation and all its triples.

        Returns:
            batch_data: The training data.
        """
        batch_data = {}
        query, train_and_test1 = data[0].keys(), data[0].values()
        query = str(list(query)).lstrip('[').rstrip(']').replace("'", "")
        #  降维度
        train_and_test1 = list(train_and_test1)
        task_triples = [item for sublist in train_and_test1 for item in sublist]

        if self.args.type_constrain:
            candidates = self.rel2candidates[query]
        else:
            candidates = list(self.ent2id.keys())
            random.shuffle(candidates)
            candidates = candidates[:1000]

        random.shuffle(task_triples)

        support_triples = task_triples[:self.args.few]
        support_pairs = [[self.symbol2id[triple[0]], self.symbol2id[triple[2]], float(triple[3])] for triple in
                         support_triples]  # (h, t, s)
        support_left = [self.ent2id[triple[0]] for triple in support_triples]
        support_right = [self.ent2id[triple[2]] for triple in support_triples]

        all_test_triples = task_triples[self.args.few:]

        if len(all_test_triples) < self.args.train_bs:
            query_triples = [random.choice(all_test_triples) for _ in range(self.args.train_bs)]
        else:
            query_triples = random.sample(all_test_triples, self.args.train_bs)

        query_pairs = [[self.symbol2id[triple[0]], self.symbol2id[triple[2]], float(triple[3])] for triple in
                       query_triples]  # (h, t, s)
        query_left = [self.ent2id[triple[0]] for triple in query_triples]
        query_right = [self.ent2id[triple[2]] for triple in query_triples]
        query_confidence = [float(triple[3]) for triple in query_triples]

        false_pairs, false_left, false_right = self.generate_false(query_triples, candidates)

        support_meta = self.get_meta(support_left, support_right)
        query_meta = self.get_meta(query_left, query_right)
        false_meta = self.get_meta(false_left, false_right)

        # symbolid-ic value pairs for loss
        symbolid_ic = []
        if self.args.has_ont:
            if self.args.rel_uc == 1:
                rel2ic = self.rel_uc1
            else:
                rel2ic = self.rel_uc2

            symbols = set()
            for triple in task_triples:
                e1 = triple[0]
                e2 = triple[2]
                if e1 not in symbols:
                    symbolid_ic.append([self.symbol2id[e1], float(self.ent2ic[e1])])
                    symbols.add(e1)
                if e2 not in symbols:
                    symbolid_ic.append([self.symbol2id[e2], float(self.ent2ic[e2])])
                    symbols.add(e2)
            symbolid_ic.append([self.symbol2id[query], float(rel2ic[query])])

        if self.args.if_GPU:
            support = Variable(torch.Tensor(support_pairs)).cuda()
            query = Variable(torch.Tensor(query_pairs)).cuda()
            false = Variable(torch.Tensor(false_pairs)).cuda()
            symbolid_ic = Variable(torch.Tensor(symbolid_ic)).cuda()
        else:
            support = Variable(torch.Tensor(support_pairs))
            query = Variable(torch.LongTensor(query_pairs))
            false = Variable(torch.LongTensor(false_pairs))
            symbolid_ic = Variable(torch.Tensor(symbolid_ic))

        batch_data["support"] = support
        batch_data["query"] = query
        batch_data["false"] = false
        batch_data["query_confidence"] = query_confidence
        batch_data["support_meta"] = support_meta
        batch_data["query_meta"] = query_meta
        batch_data["false_meta"] = false_meta
        batch_data["symbolid_ic"] = symbolid_ic

        return batch_data

    def get_meta(self, left, right):
        """get meta data
        """
        if self.args.if_GPU:
            left_connections = Variable(
                torch.Tensor(np.stack([self.connections[_, :, :] for _ in left], axis=0))).cuda()
            left_degrees = Variable(torch.FloatTensor([self.e1_degrees[_] for _ in left])).cuda()
            right_connections = Variable(
                torch.Tensor(np.stack([self.connections[_, :, :] for _ in right], axis=0))).cuda()
            right_degrees = Variable(torch.FloatTensor([self.e1_degrees[_] for _ in right])).cuda()
        else:
            left_connections = Variable(torch.LongTensor(np.stack([self.connections[_, :, :] for _ in left], axis=0)))
            left_degrees = Variable(torch.FloatTensor([self.e1_degrees[_] for _ in left]))
            right_connections = Variable(torch.LongTensor(np.stack([self.connections[_, :, :] for _ in right], axis=0)))
            right_degrees = Variable(torch.FloatTensor([self.e1_degrees[_] for _ in right]))

        return left_connections, left_degrees, right_connections, right_degrees

    def get_sampling_keys(self):
        return ['support', 'query', 'false', 'query_confidence', 'support_meta', 'query_meta', 'false_meta']


class GMUCTestSampler(object):
    """Sampling triples and recording positive triples for testing.

    Attributes:
        sampler: The function of training sampler.
        num_ent: The number of entities.
        rel2candidates: Record the entities corresponding to the same relation, type: list.
        symbol2id: Encoding the entity and relation in triples, type: dict.
        ent2id: Encoding the entity in triples, type: dict.
    """

    def __init__(self, sampler):
        self.sampler = sampler
        self.args = self.sampler.args
        self.num_ent = sampler.args.num_ent
        self.rel2candidates = sampler.rel2candidates
        # self.test_with_neg_tasks = sampler.test_with_neg_tasks
        self.symbol2id = sampler.symbol2id
        self.ent2id = sampler.ent2id

    def sampling(self, data):
        """Sampling triples and recording positive triples for testing.

        Args:
            data: The triples used to be sampled.

        Returns:
            batch_data: The data used to be evaluated.
        """
        query = str(list(data[0].keys())).replace("'", "").replace("[", "").replace("]", "")
        triple = list(data[0].values())
        triples = [item for sublist in triple for item in sublist]

        batch_data = {}
        batch_data["query"] = query
        batch_data["triples"] = triples

        return batch_data

    def get_sampling_keys(self):
        return ["query", "triples"]


class KGDataset(Dataset):

    def __init__(self, triples):
        self.triples = triples

    def __len__(self):
        return len(self.triples)

    def __getitem__(self, idx):
        return self.triples[idx]



from .BaseLitModel import BaseLitModel
from ..eval_task import *
import pickle
from ..eval_task.nDCG import mean_ndcg


class UKGELitModel(BaseLitModel):
    """Processing of training, evaluation and testing.
    """

    def __init__(self, model, args):
        super().__init__(model, args)

    def forward(self, x):
        return self.model(x)

    @staticmethod
    def add_to_argparse(parser):
        parser.add_argument("--lr", type=float, default=0.1)
        parser.add_argument("--weight_decay", type=float, default=0.01)
        return parser

    def training_step(self, batch, batch_idx):
        """Getting samples and training in KG model.
        
        Args:
            batch: The training data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            loss: The training loss for back propagation.
        """

        pos_sample = batch["positive_sample"]
        neg_sample = batch["negative_sample"]
        mode = batch["mode"]

        pos_score = self.model(pos_sample)
        neg_score = self.model(pos_sample, neg_sample, mode)

        head_emb, relation_emb, tail_emb = self.model.tri2emb(pos_sample[:, :3].to(torch.int))
        head_emb = head_emb.squeeze(1)
        relation_emb = relation_emb.squeeze(1)
        tail_emb = tail_emb.squeeze(1)
        regularizer = (torch.sum(torch.square(head_emb)) / head_emb.shape[0]) + \
                      (torch.sum(torch.square(relation_emb)) / relation_emb.shape[0]) + \
                      (torch.sum(torch.square(tail_emb)) / tail_emb.shape[0])
        loss = self.loss(pos_score, neg_score, pos_sample) + 0.005 * regularizer
        self.log("Train|loss", loss, on_step=False, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        """Getting samples and validation in KG model.
        
        Args:
            batch: The evalutaion data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            results: mrr and hits@1,3,5,10.
        """
        results = dict()

        results["count_for_conf"] = batch['positive_sample'].shape[0]
        MAE, MSE = conf_predict(batch, self.model)
        results["MAE"] = MAE.item()
        results["MSE"] = MSE.item()

        prediction = "tail"

        confidence = self.args.confidence_filter
        ranks = link_predict(batch, self.model, prediction=prediction)
        ranks_link_predict = link_predict_filter(batch, self.model, confidence, prediction=prediction)
        ranks_link_predict_raw = link_predict_raw(batch, self.model, confidence, prediction=prediction)
        results["count_for_link"] = torch.numel(ranks_link_predict)
        results["mrr"] = torch.sum(1.0 / ranks_link_predict).item()
        results["mr"] = torch.sum(ranks_link_predict).item()
        results["raw_count_for_link"] = torch.numel(ranks_link_predict_raw)
        results["raw_mrr"] = torch.sum(1.0 / ranks_link_predict_raw).item()
        results["raw_mr"] = torch.sum(ranks_link_predict_raw).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks_link_predict[ranks_link_predict <= k])
            results['raw_hits@{}'.format(k)] = torch.numel(ranks_link_predict_raw[ranks_link_predict_raw <= k])

        pos_triple = batch["positive_sample"]
        mask = pos_triple[:, -1] >= confidence
        """calculate WMR(Weighted-MR) and WMRR """
        if prediction == "all":
            conf = torch.cat([batch['positive_sample'][:, 3]] * 2)
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        else:
            conf = batch['positive_sample'][:, 3]
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        return results

    def validation_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Eval")
        with open(os.path.join(self.args.data_path, "ndcg_test.pickle"), 'rb') as f:
            hr_map = pickle.load(f)  # unpickle
        ent_num = self.args.num_ent
        ndcg, exp_ndcg = mean_ndcg(hr_map, self.model, ent_num)
        outputs["Test_ndcg"] = ndcg
        outputs["Test_exp_ndcg"] = exp_ndcg

        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        results = dict()

        results["count_for_conf"] = batch['positive_sample'].shape[0]
        MAE, MSE = conf_predict(batch, self.model)
        results["MAE"] = MAE.item()
        results["MSE"] = MSE.item()

        prediction = "tail"
        confidence = self.args.confidence_filter
        ranks = link_predict(batch, self.model, prediction=prediction)
        ranks_link_predict = link_predict_filter(batch, self.model, confidence, prediction=prediction)
        ranks_link_predict_raw = link_predict_raw(batch, self.model, confidence, prediction=prediction)
        results["count_for_link"] = torch.numel(ranks_link_predict)
        results["mrr"] = torch.sum(1.0 / ranks_link_predict).item()
        results["mr"] = torch.sum(ranks_link_predict).item()
        results["raw_count_for_link"] = torch.numel(ranks_link_predict_raw)
        results["raw_mrr"] = torch.sum(1.0 / ranks_link_predict_raw).item()
        results["raw_mr"] = torch.sum(ranks_link_predict_raw).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks_link_predict[ranks_link_predict <= k])
            results['raw_hits@{}'.format(k)] = torch.numel(ranks_link_predict_raw[ranks_link_predict_raw <= k])

        """calculate WMR(Weighted-MR) and WMRR """
        pos_triple = batch["positive_sample"]
        mask = pos_triple[:, -1] >= confidence
        if prediction == "all":
            conf = torch.cat([batch['positive_sample'][:, 3]] * 2)
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        else:
            conf = batch['positive_sample'][:, 3]
            conf_high_score = conf[mask]
            results["wmr"] = torch.sum(ranks_link_predict * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict
            results["wmrr"] = torch.sum(ranks_mrr * conf_high_score)
            results["sum_for_conf"] = torch.sum(conf_high_score)
            results["raw_wmr"] = torch.sum(ranks_link_predict_raw * conf_high_score)
            ranks_mrr = 1.0 / ranks_link_predict_raw
            results["raw_wmrr"] = torch.sum(ranks_mrr * conf_high_score)

        return results

    def test_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Test")

        with open(os.path.join(self.args.data_path, "ndcg_test.pickle"), 'rb') as f:
            hr_map = pickle.load(f)  # unpickle
        ent_num = self.args.num_ent
        ndcg, exp_ndcg = mean_ndcg(hr_map, self.model, ent_num)
        outputs["Test_ndcg"] = ndcg
        outputs["Test_exp_ndcg"] = exp_ndcg

        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def configure_optimizers(self):
        """Setting optimizer and lr_scheduler.
        Returns:
            optim_dict: Record the optimizer and lr_scheduler, type: dict.
        """
        milestones = int(self.args.max_epochs / 8)
        beta1 = 0.9  # set β1
        beta2 = 0.99  # set β2
        optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr, betas=(beta1, beta2))
        StepLR = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[milestones], gamma=0.1)
        optim_dict = {'optimizer': optimizer, 'lr_scheduler': StepLR}
        return optim_dict



import torch.nn as nn
import torch
from .model import Model
import torch.nn.functional as F
from collections import defaultdict as ddict
import csv


class UPGAT(Model):

    """`UPGAT: Uncertainty-Aware Pseudo-neighbor Augmented Knowledge GraphAttentionNetwork`_ (UPGAT), which introduces graph attention mechanism to capture the probabilistic subgraph features.

    Attributes:
        args: Model configuration parameters.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation embedding, shape:[num_rel, emb_dim].
        w: Weight when calculate confidence scores.
        b: Bias when calculate confidence scores.

    .. _UPGAT: Uncertainty-Aware Pseudo-neighbor Augmented Knowledge GraphAttentionNetwork: https://link.springer.com/chapter/10.1007/978-3-031-33377-4_5
    """

    def __init__(self, args):
        super(UPGAT, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.w = nn.Parameter(torch.tensor(1.0), requires_grad=True)
        self.b = nn.Parameter(torch.tensor(0.0), requires_grad=True)

        self.init_emb()
        self.special_spmm_final = SpecialSpmmFinal()

    def init_emb(self):
        """Initialize the model and entity and relation embeddings.

        Args:
            ent_emb_init: Entity embedding, shape:[num_ent, emb_dim].
            rel_emb_init: Relation embedding, shape:[num_rel, emb_dim].
            ent_emb: The final embedding used in the model.
            rel_emb: The final embedding used in the model.
            W_1: Weight when calculate attention scores.
            W_a: Weight when calculate attention scores.
            g_0: Embedding for the special self-loop relation of the attention baseline.
            W_E: Weight when update entity embedding.
            W_R: Weight when update relation embedding.
        """
        self.ent_emb_init = nn.Parameter(torch.randn(self.args.num_ent, self.args.emb_dim))
        self.rel_emb_init = nn.Parameter(torch.randn(self.args.num_rel, self.args.emb_dim))
        self.ent_emb = nn.Parameter(torch.randn(self.args.num_ent, self.args.emb_dim))
        self.rel_emb = nn.Parameter(torch.randn(self.args.num_rel, self.args.emb_dim))

        self.W_1 = nn.Parameter(torch.zeros(self.args.emb_dim, self.args.emb_dim))
        self.W_a = nn.Parameter(torch.zeros(1, self.args.emb_dim))
        self.g_0 = nn.Parameter(torch.zeros(1, self.args.emb_dim))
        self.W_E = nn.Parameter(torch.zeros(self.args.emb_dim, self.args.emb_dim))
        self.W_R = nn.Parameter(torch.zeros(self.args.emb_dim, self.args.emb_dim))
        nn.init.xavier_uniform_(self.g_0.data, gain=1.414)
        nn.init.xavier_uniform_(self.W_1.data, gain=1.414)
        nn.init.xavier_uniform_(self.W_a.data, gain=1.414)
        nn.init.xavier_uniform_(self.W_E.data, gain=1.414)
        nn.init.xavier_uniform_(self.W_R.data, gain=1.414)

    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        if mode == 'head-batch':
            score = head_emb * (relation_emb * tail_emb)
        else:
            score = (head_emb * relation_emb) * tail_emb
        score = score.sum(dim=-1)
        score = score.to(self.args.gpu)
        score = torch.sigmoid(self.w * score + self.b)
        return score

    def forward(self, triples, adj_matrix=None, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            adj_matrix: The adjacency matrix of the triples.
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        triples = triples[:, :3].to(torch.long)
        head_emb, relation_emb, tail_emb = self.forward_GAT(triples, adj_matrix, negs, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def forward_GAT(self, triples, adj_matrix, negs, mode):
        """The functions used in the training phase for updating embedding of triples.

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            adj_matrix: The adjacency matrix of the triples.
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            head_emb: Head entity embedding.
            relation_emb: Relation embedding.
            tail_emb: Tail entity embedding.
        """
        node_list = adj_matrix[0].to(self.args.gpu)  # ent
        edge_list = adj_matrix[1].to(self.args.gpu)  # rel
        self.ent_emb_init.data = F.normalize(self.ent_emb_init.data, p=2, dim=1).detach()
        h_i = self.ent_emb_init[node_list[0, :], :]
        h_j = self.ent_emb_init[node_list[1, :], :]
        g_k = self.rel_emb_init[edge_list]

        c_ijk = self.W_1.mm((h_j * g_k).t())
        a_ijk = -nn.LeakyReLU(0.2)(self.W_a.mm(h_i.t() * torch.tanh(c_ijk)).squeeze())

        # get attention scores
        c_baseline = self.W_1.mm((h_i * self.g_0).t())
        a_baseline = -nn.LeakyReLU(0.2)(self.W_a.mm(h_i.t() * torch.tanh(c_baseline)).squeeze())

        a_baseline_exp = torch.exp(a_baseline).unsqueeze(1)
        a_ijk_exp = torch.exp(a_ijk).unsqueeze(1)
        assert not torch.isnan(a_baseline_exp).any()
        assert not torch.isnan(a_ijk_exp).any()

        a_ijk_exp_sum = self.special_spmm_final(node_list, a_ijk_exp, self.args.num_ent, a_ijk_exp.shape[0], 1)
        a_baseline_sorted = torch.zeros_like(a_ijk_exp_sum).to(self.args.gpu)
        a_baseline_sorted[node_list[0]] = a_baseline_exp
        denominator = a_baseline_sorted + a_ijk_exp_sum
        denominator[denominator == 0.0] = 1e-12

        a_ijk_exp = a_ijk_exp.squeeze(1)
        a_ijk_exp = nn.Dropout(0.3)(a_ijk_exp)
        a_baseline_exp = a_baseline_exp.squeeze(1)
        a_baseline_exp = nn.Dropout(0.3)(a_baseline_exp)

        ac_ijk = (a_ijk_exp * c_ijk).t()
        ac_baseline = (a_baseline_exp * c_baseline).t()

        a_ijk_new = self.special_spmm_final(node_list, ac_ijk, self.args.num_ent, ac_ijk.shape[0],
                                            self.args.emb_dim)
        a_baseline_new = torch.zeros_like(a_ijk_new).to(self.args.gpu)
        a_baseline_new[node_list[0]] = ac_baseline

        assert not torch.isnan(a_baseline_new).any()
        assert not torch.isnan(a_ijk_new).any()

        h_prime = F.elu((a_baseline_new + a_ijk_new).div(denominator))
        assert not torch.isnan(h_prime).any()

        # update embeddings
        ent_emb_new = self.ent_emb_init.mm(self.W_E) + h_prime
        rel_emb_new = self.rel_emb_init.mm(self.W_R)
        ent_emb_new = F.normalize(ent_emb_new, p=2, dim=1)

        self.ent_emb.data = ent_emb_new.data
        self.rel_emb.data = rel_emb_new.data

        if mode == "single":
            head_emb = ent_emb_new[triples[:, 0]].unsqueeze(1)  # [bs, 1, dim]
            relation_emb = rel_emb_new[triples[:, 1]].unsqueeze(1)  # [bs, 1, dim]
            tail_emb = ent_emb_new[triples[:, 2]].unsqueeze(1)  # [bs, 1, dim]
        elif mode == "head-batch":
            head_emb = ent_emb_new[negs]  # [bs, num_neg, dim]
            relation_emb = rel_emb_new[triples[:, 1]].unsqueeze(1)  # [bs, 1, dim]
            tail_emb = ent_emb_new[triples[:, 2]].unsqueeze(1)  # [bs, 1, dim]
        elif mode == "tail-batch":
            head_emb = ent_emb_new[triples[:, 0]].unsqueeze(1)  # [bs, 1, dim]
            relation_emb = rel_emb_new[triples[:, 1]].unsqueeze(1)  # [bs, 1, dim]
            tail_emb = ent_emb_new[negs]

        return head_emb, relation_emb, tail_emb

    def forward_tri2emb(self, triples, mode):
        """The functions used in the testing phase for getting embedding of triples.

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            head_emb: Head entity embedding.
            relation_emb: Relation embedding.
            tail_emb: Tail entity embedding.
        """
        if mode == 'single':
            head_emb = self.ent_emb[triples[:, 0]].unsqueeze(1) # [bs, 1, dim]
            relation_emb = self.rel_emb[triples[:, 1]].unsqueeze(1) # [bs, 1, dim]
            tail_emb = self.ent_emb[triples[:, 2]].unsqueeze(1) # [bs, 1, dim]
        elif mode == "head_predict":
            head_emb = self.ent_emb.unsqueeze(0)  # [1, num_ent, dim]
            relation_emb = self.rel_emb[triples[:, 1]].unsqueeze(1)  # [bs, 1, dim]
            tail_emb = self.ent_emb[triples[:, 2]].unsqueeze(1)  # [bs, 1, dim]
        elif mode == "tail_predict":
            head_emb = self.ent_emb[triples[:, 0]].unsqueeze(1)  # [bs, 1, dim]
            relation_emb = self.rel_emb[triples[:, 1]].unsqueeze(1)  # [bs, 1, dim]
            tail_emb = self.ent_emb.unsqueeze(0)  # [1, num_ent, dim]
        return head_emb, relation_emb, tail_emb

    def get_score(self, batch, mode):
        """The functions used in the testing phase.

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        triples = batch["positive_sample"]
        triples = triples[:, :3].to(torch.long)
        triples = triples.to(self.args.gpu)

        head_emb, relation_emb, tail_emb = self.forward_tri2emb(triples, mode=mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def pseudo_tail_predict(self, data, output_file_path):
        """Predicting false-negative tail entities for all (head, relation) pairs in the training data.

        Args:
            data: All training triples.
            output_file_path: The output path of false-negative tail entities.
        """
        hr2t = ddict(set)
        pseudo_triples = []
        for h, r, t, w in data:
            hr2t[(h, r)].add(t)
        for idx, value in hr2t.items():
            idx0 = torch.tensor([idx[0]]).to(self.args.gpu)
            idx1 = torch.tensor([idx[1]]).to(self.args.gpu)
            head_emb = self.ent_emb[idx0].unsqueeze(1)
            relation_emb = self.rel_emb[idx1].unsqueeze(1)
            tail_emb = self.ent_emb.unsqueeze(0)
            score = self.score_func(head_emb, relation_emb, tail_emb, "tail_predict")
            top_scores, top_indices = torch.topk(score.flatten(), k=self.args.num_pseudo, largest=True)
            for id, tail in enumerate(top_indices):
                pred_score = top_scores[id].item()
                tail = tail.item()
                if tail not in value and pred_score != 0:
                    pseudo_triples.append(((idx[0], idx[1], tail, pred_score)))

        with open(output_file_path, 'w', newline='', encoding='utf-8') as file:
            tsv_writer = csv.writer(file, delimiter='\t')
            tsv_writer.writerows(pseudo_triples)

class SpecialSpmmFunctionFinal(torch.autograd.Function):
    """
    Special function for only sparse region backpropataion layer, similar to https://arxiv.org/abs/1710.10903
    """
    @staticmethod
    def forward(ctx, node, node_w, N, E, out_features):

        a = torch.sparse_coo_tensor(
            node, node_w, torch.Size([N, N, out_features]))
        b = torch.sparse.sum(a, dim=1)
        ctx.N = b.shape[0]
        ctx.outfeat = b.shape[1]
        ctx.E = E
        ctx.indices = a._indices()[0, :]

        return b.to_dense()

    @staticmethod
    def backward(ctx, grad_output):
        grad_values = None
        if ctx.needs_input_grad[1]:
            edge_sources = ctx.indices

            grad_values = grad_output[edge_sources]

        return None, grad_values, None, None, None

class SpecialSpmmFinal(nn.Module):
    """
    Special spmm final layer, similar to https://arxiv.org/abs/1710.10903.
    """
    def forward(self, node, node_w, N, E, out_features):
        return SpecialSpmmFunctionFinal.apply(node, node_w, N, E, out_features)



from .BaseLitModel import BaseLitModel
from ..eval_task import *


class BEUrRELitModel(BaseLitModel):
    """Processing of training, evaluation and testing.
    """

    def __init__(self, model, args):
        super().__init__(model, args)

    def forward(self, x):
        return self.model(x)

    @staticmethod
    def add_to_argparse(parser):
        parser.add_argument("--lr", type=float, default=0.1)
        parser.add_argument("--weight_decay", type=float, default=0.01)
        return parser

    def training_step(self, batch, batch_idx):
        """Getting samples and training in BEUrRE model.

        Args:
            batch: The training data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            loss: The training loss for back propagation.
        """
        pos_sample = batch["positive_sample"]
        neg_sample = batch["negative_sample"]

        self.model.true_head = batch['true_head']
        self.model.true_tail = batch['true_tail']

        loss = self.loss(self.model, pos_sample, neg_sample, self.args)
        self.log("Train|loss", loss, on_step=False, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        """Getting samples and validation in BEUrRE model.

        Args:
            batch: The evalutaion data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            results: mae, mse, wmr, wmrr, mr mrr and hits@1,3,5,10.
        """
        results = dict()
        """以MSE为指标 Early Stop"""
        results["count_for_conf"] = batch['positive_sample'].shape[0]
        MAE, MSE = conf_predict(batch, self.model)
        results["MAE"] = MAE.item()
        results["MSE"] = MSE.item()

        prediction = "tail"
        """以MRR为指标 Early Stop"""
        ranks = link_predict(batch, self.model, prediction=prediction)
        results["count_for_link"] = torch.numel(ranks)
        results["mrr"] = torch.sum(1.0 / ranks).item()
        results["mr"] = torch.sum(ranks).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks[ranks <= k])

        """计算WMR   Weighted-MR """
        if prediction == "all":  # 头尾实体预测平均
            conf = torch.cat([batch['positive_sample'][:, 3]] * 2)
            results["wmr"] = torch.sum(ranks * conf)
            ranks_mrr = 1.0 / ranks
            results["wmrr"] = torch.sum(ranks_mrr * conf)
            results["sum_for_conf"] = torch.sum(conf)
        else: # 头尾实体预测
            conf = batch['positive_sample'][:, 3]
            results["wmr"] = torch.sum(ranks * conf)
            ranks_mrr = 1.0 / ranks
            results["wmrr"] = torch.sum(ranks_mrr * conf)
            results["sum_for_conf"] = torch.sum(conf)
        return results

    def validation_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Eval")

        # with open(os.path.join(self.args.data_path, "ndcg_test.pickle"), 'rb') as f:
        #     hr_map = pickle.load(f)  # unpickle
        # ent_num = self.args.num_ent
        # ndcg, exp_ndcg = mean_ndcg(hr_map, self.model, ent_num)
        # outputs["Test_ndcg"] = ndcg
        # outputs["Test_exp_ndcg"] = exp_ndcg

        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        """Getting samples and test in BEUrRE model.

        Args:
            batch: The evaluation data.
            batch_idx: The dict_key in batch, type: list.

        Returns:
            results: mae, mse, wmr, wmrr, mr mrr and hits@1,3,5,10.
        """
        results = dict()

        """以MSE为指标 Early Stop"""
        results["count_for_conf"] = batch['positive_sample'].shape[0]
        MAE, MSE = conf_predict(batch, self.model)
        results["MAE"] = MAE.item()
        results["MSE"] = MSE.item()

        prediction = "tail"
        """以MRR为指标 Early Stop"""
        ranks = link_predict(batch, self.model, prediction=prediction)
        results["count_for_link"] = torch.numel(ranks)
        results["mrr"] = torch.sum(1.0 / ranks).item()
        results["mr"] = torch.sum(ranks).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks[ranks <= k])

        """计算WMR   Weighted-MR """
        if prediction == "all":  # 头尾实体预测平均
            conf = torch.cat([batch['positive_sample'][:, 3]] * 2)
            results["wmr"] = torch.sum(ranks * conf)
            ranks_mrr = 1.0 / ranks
            results["wmrr"] = torch.sum(ranks_mrr * conf)
            results["sum_for_conf"] = torch.sum(conf)
        else: # 头尾实体预测
            conf = batch['positive_sample'][:, 3]
            results["wmr"] = torch.sum(ranks * conf)
            ranks_mrr = 1.0 / ranks
            results["wmrr"] = torch.sum(ranks_mrr * conf)
            results["sum_for_conf"] = torch.sum(conf)

        return results

    def test_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Test")

        # with open(os.path.join(self.args.data_path, "ndcg_test.pickle"), 'rb') as f:
        #     hr_map = pickle.load(f)  # unpickle
        # ent_num = self.args.num_ent
        # ndcg, exp_ndcg = mean_ndcg(hr_map, self.model, ent_num)
        # outputs["Test_ndcg"] = ndcg
        # outputs["Test_exp_ndcg"] = exp_ndcg

        self.log_dict(outputs, prog_bar=True, on_epoch=True)


    def configure_optimizers(self):
        """Setting optimizer and lr_scheduler.

        Returns:
            optim_dict: Record the optimizer and lr_scheduler, type: dict.
        """
        optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr)
        # optim_dict = {'optimizer': optimizer, 'lr_scheduler': StepLR}
        optim_dict = {'optimizer': optimizer}
        return optim_dict



import torch
import torch.nn.functional as F
import torch.nn as nn


class PASSLEAF_Loss(nn.Module):

    def __init__(self, args, model):
        super(PASSLEAF_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, pos_score, neg_score, pos_sample, semi_score = None ,semi_sample = None):

        confidence = pos_sample[:, 3]  # get confidence of positive triples
        pos_score = pos_score.squeeze()

        if semi_score is not None:
            semi_score = semi_score.squeeze()

        if semi_sample is not None:
            confidence_semi = semi_sample[:, 3] # get confidence of semi triples (the score was stored in the pool)

        loss_1 = torch.sum((pos_score - confidence) ** 2) # L_pos
        loss_2 = torch.sum(neg_score ** 2) # L_neg
        # loss is different when using semi-supervised learning
        if semi_score is not None:
            loss_3 = torch.sum((semi_score - confidence_semi) ** 2) # L_semi
            loss = loss_1 + (loss_2 + loss_3) / self.args.num_neg # L = L_pos + (L_semi + L_neg) / N_gen
        else:
            loss = loss_1 + loss_2 / neg_score.shape[1]

        return loss
